# Ignite WandB custom logger
#
# This file is derived from PyTorch Ignite (BSD 3-Clause License)
# Copyright (c) 2018-present, PyTorch-Ignite team
# License for original code: BSD 3-Clause
#
# Modifications:
# Author: Giovanni Spadaro - https://giovannispadaro.it
# Project: https://github.com/Giovo17/tfs-mt
# Documentation: https://giovo17.github.io/tfs-mt
#
# Copyright (c) Giovanni Spadaro.
# License for modifications: MIT
#
# The full text of both licenses can be found in the LICENSE file in the root directory of this source tree.

from collections.abc import Callable
from typing import Any

import torch.nn as nn
from ignite.engine import Engine, Events
from ignite.handlers.base_logger import BaseLogger, BaseOptimizerParamsHandler, BaseOutputHandler
from ignite.handlers.utils import global_step_from_engine
from torch.optim import Optimizer

__all__ = ["OptimizerParamsHandler", "OutputHandler", "WandBLogger", "global_step_from_engine"]


class WandBLogger(BaseLogger):
    """`Weights & Biases <https://wandb.ai/site>`_ handler to log metrics, model/optimizer parameters, gradients
    during training and validation. It can also be used to log model checkpoints to the Weights & Biases cloud.

    .. code-block:: bash

        pip install wandb

    This class is also a wrapper for the wandb module. This means that you can call any wandb function using
    this wrapper. See examples on how to save model parameters and gradients.

    Args:
        args: Positional arguments accepted by `wandb.init`.
        kwargs: Keyword arguments accepted by `wandb.init`.
            Please see `wandb.init <https://docs.wandb.ai/ref/python/init>`_ for documentation of possible parameters.

    Examples:
        .. code-block:: python

            from ignite.handlers.wandb_logger import *

            # Create a logger. All parameters are optional. See documentation
            # on wandb.init for details.

            wandb_logger = WandBLogger(
                entity="shared",
                project="pytorch-ignite-integration",
                name="cnn-mnist",
                config={"max_epochs": 10},
                tags=["pytorch-ignite", "minst"]
            )

            # Attach the logger to the trainer to log training loss at each iteration
            wandb_logger.attach_output_handler(
                trainer,
                event_name=Events.ITERATION_COMPLETED,
                tag="training",
                output_transform=lambda loss: {"loss": loss}
            )

            # Attach the logger to the evaluator on the training dataset and log NLL, Accuracy metrics after each epoch
            # We setup `global_step_transform=lambda *_: trainer.state.iteration` to take iteration value
            # of the `trainer`:
            wandb_logger.attach_output_handler(
                train_evaluator,
                event_name=Events.EPOCH_COMPLETED,
                tag="training",
                metric_names=["nll", "accuracy"],
                global_step_transform=lambda *_: trainer.state.iteration,
            )

            # Attach the logger to the evaluator on the validation dataset and log NLL, Accuracy metrics after
            # each epoch. We setup `global_step_transform=lambda *_: trainer.state.iteration` to take iteration value
            # of the `trainer` instead of `evaluator`.
            wandb_logger.attach_output_handler(
                evaluator,
                event_name=Events.EPOCH_COMPLETED,
                tag="validation",
                metric_names=["nll", "accuracy"],
                global_step_transform=lambda *_: trainer.state.iteration,
            )

            # Attach the logger to the trainer to log optimizer's parameters, e.g. learning rate at each iteration
            wandb_logger.attach_opt_params_handler(
                trainer,
                event_name=Events.ITERATION_STARTED,
                optimizer=optimizer,
                param_name='lr'  # optional
            )

            # We need to close the logger when we are done
            wandb_logger.close()

        If you want to log model gradients, the model call graph, etc., use the logger as wrapper of wandb. Refer
        to the documentation of wandb.watch for details:

        .. code-block:: python

            wandb_logger = WandBLogger(
                entity="shared",
                project="pytorch-ignite-integration",
                name="cnn-mnist",
                config={"max_epochs": 10},
                tags=["pytorch-ignite", "minst"]
            )

            model = torch.nn.Sequential(...)
            wandb_logger.watch(model)

        For model checkpointing, Weights & Biases creates a local run dir, and automatically synchronizes all
        files saved there at the end of the run. You can just use the `wandb_logger.run.dir` as path for the
        `ModelCheckpoint`:

        .. code-block:: python

            from ignite.handlers import ModelCheckpoint

            def score_function(engine):
                return engine.state.metrics['accuracy']

            model_checkpoint = ModelCheckpoint(
                wandb_logger.run.dir, n_saved=2, filename_prefix='best',
                require_empty=False, score_function=score_function,
                score_name="validation_accuracy",
                global_step_transform=global_step_from_engine(trainer)
            )
            evaluator.add_event_handler(Events.COMPLETED, model_checkpoint, {'model': model})

    Note:
        :class:`~ignite.handlers.wandb_logger.OutputHandler` can handle
        metrics, state attributes and engine output values of the following format:
        - scalar values (i.e. int, float)
        - 0d and 1d pytorch tensors
        - dicts and list/tuples of previous types

    """

    def __init__(self, *args: Any, **kwargs: Any):
        try:
            import wandb

            self._wandb = wandb
        except ImportError as err:
            raise ModuleNotFoundError(
                "This contrib module requires wandb to be installed. "
                "You man install wandb with the command:\n pip install wandb\n"
            ) from err
        if kwargs.get("init", True):
            kwargs.pop("init", None)
            self._wandb.init(*args, **kwargs)

    def __getattr__(self, attr: Any) -> Any:
        return getattr(self._wandb, attr)

    def watch(self, model: nn.Module) -> None:
        self._wandb.watch(model)

    def close(self) -> None:
        self._wandb.finish()

    def _create_output_handler(self, *args: Any, **kwargs: Any) -> "OutputHandler":
        return OutputHandler(*args, **kwargs)

    def _create_opt_params_handler(self, *args: Any, **kwargs: Any) -> "OptimizerParamsHandler":
        return OptimizerParamsHandler(*args, **kwargs)


class OutputHandler(BaseOutputHandler):
    """Helper handler to log engine's output and/or metrics

    Args:
        tag: common title for all produced plots. For example, "training"
        metric_names: list of metric names to plot or a string "all" to plot all available
            metrics.
        output_transform: output transform function to prepare `engine.state.output` as a number.
            For example, `output_transform = lambda output: output`
            This function can also return a dictionary, e.g `{"loss": loss1, "another_loss": loss2}` to label the plot
            with corresponding keys.
        global_step_transform: global step transform function to output a desired global step.
            Input of the function is `(engine, event_name)`. Output of function should be an integer.
            Default is None, global_step based on attached engine. If provided,
            uses function output as global_step. To setup global step from another engine, please use
            :meth:`~ignite.handlers.wandb_logger.global_step_from_engine`.
        sync: If set to False, process calls to log in a seperate thread. Default (None) uses whatever
            the default value of wandb.log.

    Examples:
        .. code-block:: python

            from ignite.handlers.wandb_logger import *

            # Create a logger. All parameters are optional. See documentation
            # on wandb.init for details.

            wandb_logger = WandBLogger(
                entity="shared",
                project="pytorch-ignite-integration",
                name="cnn-mnist",
                config={"max_epochs": 10},
                tags=["pytorch-ignite", "minst"]
            )

            # Attach the logger to the evaluator on the validation dataset and log NLL, Accuracy metrics after
            # each epoch. We setup `global_step_transform=lambda *_: trainer.state.iteration,` to take iteration value
            # of the `trainer`:
            wandb_logger.attach(
                evaluator,
                log_handler=OutputHandler(
                    tag="validation",
                    metric_names=["nll", "accuracy"],
                    global_step_transform=lambda *_: trainer.state.iteration,
                ),
                event_name=Events.EPOCH_COMPLETED
            )
            # or equivalently
            wandb_logger.attach_output_handler(
                evaluator,
                event_name=Events.EPOCH_COMPLETED,
                tag="validation",
                metric_names=["nll", "accuracy"],
                global_step_transform=lambda *_: trainer.state.iteration,
            )

        Another example, where model is evaluated every 500 iterations:

        .. code-block:: python

            from ignite.handlers.wandb_logger import *

            @trainer.on(Events.ITERATION_COMPLETED(every=500))
            def evaluate(engine):
                evaluator.run(validation_set, max_epochs=1)

            # Create a logger. All parameters are optional. See documentation
            # on wandb.init for details.

            wandb_logger = WandBLogger(
                entity="shared",
                project="pytorch-ignite-integration",
                name="cnn-mnist",
                config={"max_epochs": 10},
                tags=["pytorch-ignite", "minst"]
            )

            def global_step_transform(*args, **kwargs):
                return trainer.state.iteration

            # Attach the logger to the evaluator on the validation dataset and log NLL, Accuracy metrics after
            # every 500 iterations. Since evaluator engine does not have access to the training iteration, we
            # provide a global_step_transform to return the trainer.state.iteration for the global_step, each time
            # evaluator metrics are plotted on Weights & Biases.

            wandb_logger.attach_output_handler(
                evaluator,
                event_name=Events.EPOCH_COMPLETED,
                tag="validation",
                metrics=["nll", "accuracy"],
                global_step_transform=global_step_transform
            )

        Another example where the State Attributes ``trainer.state.alpha`` and ``trainer.state.beta``
        are also logged along with the NLL and Accuracy after each iteration:

        .. code-block:: python

            wandb_logger.attach_output_handler(
                trainer,
                event_name=Events.ITERATION_COMPLETED,
                tag="training",
                metrics=["nll", "accuracy"],
                state_attributes=["alpha", "beta"],
            )


        Example of `global_step_transform`:

        .. code-block:: python

            def global_step_transform(engine, event_name):
                return engine.state.get_event_attrib_value(event_name)

    .. versionchanged:: 0.4.7
        accepts an optional list of `state_attributes`
    """

    def __init__(
        self,
        tag: str,
        metric_names: list[str] | None = None,
        output_transform: Callable | None = None,
        global_step_transform: Callable[[Engine, str | Events], int] | None = None,
        state_attributes: list[str] | None = None,
    ):
        super().__init__(tag, metric_names, output_transform, global_step_transform, state_attributes)

    def __call__(self, engine: Engine, logger: WandBLogger, event_name: str | Events) -> None:
        if not isinstance(logger, WandBLogger):
            raise TypeError(f"Handler '{self.__class__.__name__}' works only with WandBLogger.")

        global_step = self.global_step_transform(engine, event_name)
        if not isinstance(global_step, int):
            raise TypeError(
                f"global_step must be int, got {type(global_step)}. Please check the output of global_step_transform."
            )

        metrics = self._setup_output_metrics_state_attrs(engine, log_text=True, key_tuple=False)
        logger.log(metrics, step=global_step)


class OptimizerParamsHandler(BaseOptimizerParamsHandler):
    """Helper handler to log optimizer parameters

    Args:
        optimizer: torch optimizer or any object with attribute ``param_groups``
            as a sequence.
        param_name: parameter name
        tag: common title for all produced plots. For example, "generator"
        sync: If set to False, process calls to log in a seperate thread. Default (None) uses whatever
            the default value of wandb.log.

    Examples:
        .. code-block:: python

            from ignite.handlers.wandb_logger import *

            # Create a logger. All parameters are optional. See documentation
            # on wandb.init for details.

            wandb_logger = WandBLogger(
                entity="shared",
                project="pytorch-ignite-integration",
                name="cnn-mnist",
                config={"max_epochs": 10},
                tags=["pytorch-ignite", "minst"]
            )

            # Attach the logger to the trainer to log optimizer's parameters, e.g. learning rate at each iteration
            wandb_logger.attach(
                trainer,
                log_handler=OptimizerParamsHandler(optimizer),
                event_name=Events.ITERATION_STARTED
            )
            # or equivalently
            wandb_logger.attach_opt_params_handler(
                trainer,
                event_name=Events.ITERATION_STARTED,
                optimizer=optimizer
            )
    """

    def __init__(self, optimizer: Optimizer, param_name: str = "lr", tag: str | None = None, sync: bool | None = None):
        super().__init__(optimizer, param_name, tag)

    def __call__(self, engine: Engine, logger: WandBLogger, event_name: str | Events) -> None:
        if not isinstance(logger, WandBLogger):
            raise TypeError("Handler OptimizerParamsHandler works only with WandBLogger")

        global_step = engine.state.get_event_attrib_value(event_name)
        tag_prefix = f"{self.tag}/" if self.tag else ""
        params = {
            f"{tag_prefix}{self.param_name}/group_{i}": float(param_group[self.param_name])
            for i, param_group in enumerate(self.optimizer.param_groups)
        }
        logger.log(params, step=global_step)



# Transformer data utils
#
# Author: Giovanni Spadaro - https://giovannispadaro.it
# Project: https://github.com/Giovo17/tfs-mt
# Documentation: https://giovo17.github.io/tfs-mt
#
# Copyright (c) Giovanni Spadaro.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in the root directory of this source tree.

import json
import os
import re
import string
import typing
import zipfile
from abc import ABC, abstractmethod
from collections import Counter
from functools import partial
from itertools import chain

import numpy as np
import requests
import torch
import torch.nn.functional as F
from datasets import load_dataset
from omegaconf.dictconfig import DictConfig
from omegaconf.listconfig import ListConfig
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import DataLoader, Dataset


class VocabNotBuiltError(Exception):
    def __init__(self, msg="Vocabulary not built. Call build_vocab first."):
        super().__init__(msg)


class CheckpointEmptyVocabException(Exception):
    def __init__(self, msg="The provided json file has an empty vocab."):
        super().__init__(msg)


class GloVeVersionError(Exception):
    def __init__(self, glove_version, glove_available_versions):
        msg = f"GloVe version is not available, got {glove_version}, available versions: {glove_available_versions}."
        super().__init__(msg)


class DatasetLanguagesNotAvailableError(Exception):
    def __init__(self, src_lang, tgt_lang):
        msg = (
            f"Choosen languages are not available in dataset, got src_lang = '{src_lang}' and tgt_lang = '{tgt_lang}'."
        )
        super().__init__(msg)


class BaseTokenizer(ABC):
    @property
    @abstractmethod
    def vocab_size(self):
        pass

    @property
    @abstractmethod
    def sos_token_idx(self):
        pass

    @property
    @abstractmethod
    def eos_token_idx(self):
        pass

    @property
    @abstractmethod
    def pad_token_idx(self):
        pass

    @property
    @abstractmethod
    def unk_token_idx(self):
        pass

    @abstractmethod
    def build_vocab(self, *args, **kwargs) -> None:
        pass

    @classmethod
    @abstractmethod
    def from_pretrained(cls, path: str):
        pass

    @abstractmethod
    def to_json(self, path: str) -> None:
        pass

    @abstractmethod
    def tokenize(self, text: str) -> list[str]:
        pass

    @abstractmethod
    def encode(
        self,
        input_sequence: str | list[str] | np.ndarray | torch.Tensor,
        pad_to_len: int | None = None,
        return_only_mask: bool = False,
    ) -> tuple[np.ndarray, np.ndarray]:
        pass

    @abstractmethod
    def decode(self, token_ids: np.ndarray | list[int]) -> list[str]:
        pass


def download_glove(output_dir: str, glove_version: str = "glove.2024.wikigiga.50d") -> str:
    """Download GloVe embeddings and returns the filepath."""
    glove_folder_path = output_dir + f"/{glove_version}"
    os.makedirs(glove_folder_path, exist_ok=True)

    url = f"https://nlp.stanford.edu/data/wordvecs/{glove_version}.zip"
    zip_path = output_dir + f"/{glove_version}.zip"

    glove_filepath = None
    for file in os.listdir(glove_folder_path):
        if file.endswith(".txt"):
            glove_filepath = os.path.join(glove_folder_path, file)
            break

    if glove_filepath is None:
        print(f"GloVe not found in {glove_folder_path}. Downloading GloVe ({glove_version})...")

        response = requests.get(url, stream=True, timeout=600)
        response.raise_for_status()

        with open(zip_path, "wb") as f:
            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)

        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            zip_ref.extractall(glove_folder_path)
        os.remove(zip_path)

        for file in os.listdir(glove_folder_path):
            if file.endswith(".txt"):
                glove_filepath = os.path.join(glove_folder_path, file)
                break

    return glove_filepath


class WordTokenizer(BaseTokenizer):
    """Word tokenizer.
    Mainly used to let the model be compatible with pretrained GloVe embeddings.

    Args:
        special_tokens (dict[str, str] | None, optional): Special tokens to be considered, eg. SOS_TOKEN, EOS_TOKEN. Defaults to None.
        contractions (list[str] | None, optional): Contractions to be considered, eg. 's, 'll. If None the following set of contractions will be considered: `'s`, `'re`, `'ve`, `'m`, `'ll`, `'d`, `n't`, `'t`, `n'ts`. Defaults to None.
        tokenizer_max_len (int, optional): Tokenizer max sequence length. Mainly used to limit memory usage and performance impact during training and inference due to attention quadratic complexity. Defaults to 128.
        max_vocab_size (int, optional): Maximum number of token in vocabulary. Defaults to 100_000.
    """

    def __init__(
        self,
        special_tokens: dict[str, str] | None = None,
        contractions: list[str] | None = None,
        tokenizer_max_len: int = 128,
        max_vocab_size: int = 100_000,
    ):
        self.vocab: dict[str, int] = {}
        self.vocab_reverse: dict[int, str] = {}  # Useful for efficient decoding

        self.tokenizer_max_len = tokenizer_max_len
        self.max_vocab_size = max_vocab_size

        self.special_tokens = special_tokens or {
            "sos_token": "<s>",
            "eos_token": "</s>",
            "pad_token": "<PAD>",
            "unk_token": "<UNK>",
        }

        # Setup contractions management utilities
        contractions_init = contractions or ["'s", "'re", "'ve", "'m", "'ll", "'d", "n't", "'t", "n'ts"]

        # Prepare a dict of the following type to address the detection of nested contractions:
        # nested_contractions_dict = {
        #     "ANYTHING_n't_ANYTHING": ["n't"],
        #     "n't": ["'t"]
        # }
        # NOTE "'t" is technically a sub contraction of both "ANYTHING_n't_ANYTHING" and "n't" but it will be only considered as sub contraction of n't
        self.nested_contractions_dict: dict[str, list[str]] = {}

        for contraction in contractions_init:
            other_contractions = contractions_init.copy()
            other_contractions.remove(contraction)
            for comp_contraction in other_contractions:
                if contraction in comp_contraction:  # Detect sub contraction
                    # Create new mapping in dict
                    if comp_contraction not in self.nested_contractions_dict:
                        self.nested_contractions_dict[comp_contraction] = [contraction]
                    # Detect sub contraction already present in the dict
                    elif comp_contraction in self.nested_contractions_dict and contraction not in list(
                        chain.from_iterable(self.nested_contractions_dict.values())
                    ):
                        # Merge all elements in sub contractions lists in a single list
                        self.nested_contractions_dict[comp_contraction].append(contraction)

        # Final sets of sub contractions and non-nested contractions
        # NOTE the plain list of all sub contractions will be used during tokenizationin order to better detect correct splitting when the `'` is encountered
        self.all_sub_contractions = list(chain.from_iterable(self.nested_contractions_dict.values()))
        self.contractions = [c for c in contractions_init if c not in self.all_sub_contractions]

        self.glove_available_versions = [
            "glove.2024.dolma.300d",
            "glove.2024.wikigiga.300d",
            "glove.2024.wikigiga.200d",
            "glove.2024.wikigiga.100d",
            "glove.2024.wikigiga.50d",
            "glove.42B.300d",
            "glove.6B",
            "glove.840B.300d",
            "glove.twitter.27B",
        ]

    @property
    def vocab_size(self):
        return len(self.vocab.items()) if self.vocab else 0

    @property
    def sos_token_idx(self):
        return self.vocab.get(self.special_tokens["sos_token"], 0)

    @property
    def eos_token_idx(self):
        return self.vocab.get(self.special_tokens["eos_token"], 1)

    @property
    def pad_token_idx(self):
        return self.vocab.get(self.special_tokens["pad_token"], 2)

    @property
    def unk_token_idx(self):
        return self.vocab.get(self.special_tokens["unk_token"], 3)

    @classmethod
    def from_pretrained(cls: type["WordTokenizer"], path: str) -> "WordTokenizer":
        if path.startswith(("https://", "http://")):
            try:
                response = requests.get(path, timeout=100)
                response.raise_for_status()
                data = response.json()
            except Exception as e:
                raise RuntimeError(f"Failed to download tokenizer from {path}: {e}") from e
        else:
            if os.path.exists(path):
                with open(path, encoding="utf-8") as f:
                    data = json.load(f)

        if data.get("vocab") is None:
            raise CheckpointEmptyVocabException()

        tokenizer = cls(special_tokens=data.get("special_tokens", {}), contractions=data.get("contractions", {}))
        tokenizer.vocab = data.get("vocab")
        tokenizer.vocab_reverse = {idx: token for token, idx in tokenizer.vocab.items()}

        return tokenizer

    def to_json(self, path: str) -> None:
        if self.vocab_size == 0:
            raise VocabNotBuiltError()

        to_save = {
            "type": "WordTokenizer",
            "special_tokens": self.special_tokens,
            "contractions": self.contractions,
            "vocab": self.vocab,
        }

        with open(path, "w", encoding="utf-8") as f:
            json.dump(to_save, f, ensure_ascii=False, indent=2)

    def build_vocab(
        self,
        tokens: list[str],
        min_freq: int = 2,
        extend_with_glove: bool = False,
        glove_version: str = "glove.2024.wikigiga.50d",
        **kwargs,
    ) -> None:
        """Build vocabulary method.

        Args:
            tokens (list[str]): Tokens from dataset to build vocabulary on.
            min_freq (int, optional): Minimum number of times a token has to appear in the dataset to be included in the vocabulary. Defaults to 2.
            extend_with_glove (bool, optional): Enable vocabulary extension with GloVe tokens. Defaults to False.
            glove_version (str, optional): GloVe version to use if `extend_with_glove` is `True`. Defaults to "glove.2024.wikigiga.50d".

        Raises:
            GloVeVersionError: Raised when supplied glove_version is unavailable.
        """
        vocab_list = []
        vocab_list.extend(self.special_tokens.values())
        vocab_set = set(vocab_list)

        if min_freq > 1:
            token_counts = Counter(tokens)
            for token, count in token_counts.items():
                # len(vocab_set) is O(1) operation cause the length is stored as a set attribute
                if (count >= min_freq and self.max_vocab_size == -1) or (
                    count >= min_freq and self.max_vocab_size > 0 and len(vocab_set) < self.max_vocab_size
                ):
                    vocab_set.add(token.lower())
        else:
            for token in tokens:
                if self.max_vocab_size == -1 or (self.max_vocab_size > 0 and len(vocab_set) < self.max_vocab_size):
                    vocab_set.add(token.lower())

        vocab_list = list(vocab_set)
        del vocab_set

        if extend_with_glove and (
            self.max_vocab_size == -1 or (self.max_vocab_size > 0 and len(vocab_list) < self.max_vocab_size)
        ):
            print("Extending vocab with GloVe tokens...")

            if glove_version not in self.glove_available_versions:
                raise GloVeVersionError(glove_version, self.glove_available_versions)

            data_path = os.getcwd() + "/data" if "data_path" not in kwargs else kwargs["data_path"]

            glove_tokens = []

            try:
                glove_filepath = download_glove(data_path, glove_version)

                print(f"Loading GloVe {glove_version} tokens from file...")

                with open(glove_filepath, encoding="utf-8") as f:
                    lines = f.readlines()

                # Parse GloVe tokens
                for line in lines:
                    parts = line.strip().split()
                    try:
                        float(parts[1])
                    except ValueError:
                        continue
                    else:
                        token = parts[0].lower()
                        glove_tokens.append(token)

                initial_size = len(vocab_list)
                if self.max_vocab_size != -1:
                    for tok in glove_tokens:
                        if tok not in vocab_list:
                            vocab_list.append(tok)
                        if len(vocab_list) == self.max_vocab_size:
                            break
                else:
                    vocab_list.extend(glove_tokens)
                    vocab_list = list(set(vocab_list))

                print(f"Added {len(vocab_list) - initial_size} tokens from GloVe")

            except Exception as e:
                print(f"Error with GloVe processing GloVe: {e}")

        # Create mappings
        self.vocab = {token: idx for idx, token in enumerate(vocab_list)}
        self.vocab_reverse = dict(enumerate(vocab_list))

        del vocab_list

        print(f"Built vocabulary with {len(self.vocab.items())} tokens.")

    def tokenize(self, text: str) -> list[str]:
        """Tokenizer based on GloVe word tokenizer in order to let the model be compatible with GloVe pretrained embeddings.

        Note:
            Max word length is 1000. Contractions are treated as distinct tokens, eg. `n't`, `'s`, `'ll`.

        Note: Reference
            GloVe tokenizer source code available [here](https://github.com/stanfordnlp/GloVe/blob/master/src/common.c#L75)

        Args:
            text (str): text to be tokenized.

        Returns:
            list[str]: List of string tokens from text.
        """

        text = text.strip().lower()

        for contraction in self.contractions:
            pattern = r"([a-zA-Z]+)" + re.escape(contraction) + r"\b"
            text = re.sub(pattern, r"\1 " + contraction, text)

        for complete_contraction, sub_list in self.nested_contractions_dict.items():
            pattern = r"([a-zA-Z]+)" + re.escape(complete_contraction) + r"\b"
            if not re.compile(complete_contraction).search(text):
                for sub_contraction in sub_list:
                    pattern = r"([a-zA-Z]+)" + re.escape(sub_contraction) + r"\b"
                    text = re.sub(pattern, r"\1 " + sub_contraction, text)
            else:
                text = re.sub(pattern, r"\1 " + complete_contraction, text)

        words = text.split()

        tokens = []
        for word in words:
            # Split words when encountering a `'` that's not involved in a quote or a contraction.
            # eg. "Quell'ultimo" gets splitted into "Quell'" and "ultimo"
            if (
                "'" in word
                and word[0] != "'"
                and word[-1] != "'"
                and word not in self.contractions
                and word not in self.all_sub_contractions
            ):
                parts = word.split("'")
                for i, part in enumerate(parts):
                    if part:
                        if i < len(parts) - 1:
                            part += "'"
                        tokens.append(part)

            # Handle trailing punctuation
            elif word and word[-1] in string.punctuation and word[-1] != "'":
                # Strip all trailing punctuation
                core_word = word.rstrip(string.punctuation)
                punct = word[len(core_word) :]
                if core_word:
                    tokens.append(core_word)
                for p in punct:
                    tokens.append(p)

            else:
                if word:
                    tokens.append(word)

        # Truncate sequences by default
        return tokens[: self.tokenizer_max_len]

    def encode(
        self,
        input_sequence: str | list[str] | np.ndarray | torch.Tensor,
        pad_to_len: int | None = None,
        return_only_mask: bool = False,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Tokenizer encode function.

        It also returns the mask to be used during attention in order not compute it with respect to `PAD` tokens.

        Note:
            The mask is designed to be True where there's a token with the model has to compute attention to, False otherwise.

        Args:
            input_sequence (str | list[str] | np.ndarray | torch.Tensor): Sequence to be encoded or already encoded (useful in decoding stage when this method is only used to provide the mask).
            pad_to_len (int | None, optional): Sequence length to pad `input_sequence` with. Defaults to None.
            return_only_mask (bool, optional): Return only attention mask. Defaults to False.

        Raises:
            VocabNotBuiltError: Raised when vocabulary is not built.

        Returns:
            tuple[np.ndarray, np.ndarray]: Array of token ids and mask.
        """

        if self.vocab_size == 0:
            raise VocabNotBuiltError()

        if return_only_mask:
            if isinstance(input_sequence, np.ndarray):
                mask = input_sequence != self.pad_token_idx
            elif isinstance(input_sequence, torch.Tensor):
                mask = (input_sequence != self.pad_token_idx).to(input_sequence.device)
            else:
                raise ValueError()
            return mask

        # Useful when building a TranslationDataset in order to execute tokenize method only once
        tokens = self.tokenize(input_sequence) if isinstance(input_sequence, str) else input_sequence

        # Add SOS and EOS tokens to given sequence
        if tokens[0] != self.special_tokens["sos_token"]:
            tokens.insert(0, self.special_tokens["sos_token"])
        if tokens[-1] != self.special_tokens["eos_token"]:
            tokens.append(self.special_tokens["eos_token"])

        token_ids = [self.vocab.get(token, self.unk_token_idx) for token in tokens]

        if pad_to_len is not None:  # Pad sequence to pad_to_len
            pad_to_len += 2  # Considering SOS and EOS tokens
            token_ids.extend([self.pad_token_idx for _ in range(pad_to_len - len(tokens))])

        token_ids = np.array(token_ids, dtype=np.long)

        # Mask to disable attention to pad tokens
        mask = np.array([token != self.pad_token_idx for token in token_ids], dtype=np.bool)

        return token_ids, mask

    def decode(self, token_ids: np.ndarray | list[int]) -> list[str]:
        """Decode token IDs.
        Returns the unknown token if the input token is not present in the vocabulary.

        Args:
            token_ids (np.ndarray | list[str]): Array or list of tokens ids to decode into text.

        Raises:
            VocabNotBuiltError: Vocabulary is not built.

        Returns:
            list[str]: Decoded text.
        """
        if self.vocab_size == 0:
            raise VocabNotBuiltError()
        return [self.vocab_reverse.get(idx, self.special_tokens["unk_token"]) for idx in token_ids]


class BPETokenizer(BaseTokenizer):
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        pass

    def build_vocab(self, *args, **kwargs) -> None:
        pass


class TranslationDataset(Dataset):
    """Translation Dataset.

    Args:
        src_texts (list[str]): List of source texts.
        tgt_texts (list[str]): List of target texts.
        src_tokenizer (BaseTokenizer): Tokenizer used to preprocess the source language text.
        tgt_tokenizer (BaseTokenizer): Tokenizer used to preprocess the target language text.
        src_lang (str): Identifier for the source language, e.g., `"en"` for English.
        tgt_lang (str): Identifier for the target language, e.g., `"it"` for Italian.
        max_sequence_length (int | None, optional): Maximum sequence length for tokenization. If None, sequences are not truncated. Defaults to None.
    """

    def __init__(
        self,
        src_texts: list[str],
        tgt_texts: list[str],
        src_tokenizer: BaseTokenizer,
        tgt_tokenizer: BaseTokenizer,
        src_lang: str,
        tgt_lang: str,
        max_sequence_length: int | None = None,
        **kwargs,
    ):
        self.src_tokenizer = src_tokenizer
        self.tgt_tokenizer = tgt_tokenizer
        self.src_lang = src_lang
        self.tgt_lang = tgt_lang
        self.max_sequence_length = max_sequence_length

        # Build tokenizers vocab if empty
        if self.src_tokenizer.vocab_size == 0 or self.tgt_tokenizer.vocab_size == 0:
            if "extend_vocab_with_glove" in kwargs and "glove_version" in kwargs:
                self._build_vocabs(
                    vocab_min_freq=kwargs.get("vocab_min_freq", 2),
                    extend_with_glove=kwargs.get("extend_vocab_with_glove", True),
                    glove_version=kwargs.get("glove_version", "glove.2024.wikigiga.50d"),
                )
            else:
                self._build_vocabs(kwargs.get("vocab_min_freq", 2))

        # Filter input data excluding texts longer than max_sequence_length
        if max_sequence_length is not None and max_sequence_length > 0:
            print(f"Max sequence length set to {max_sequence_length}.")
            self.src_texts, self.tgt_texts = [], []
        else:
            self.src_texts = src_texts
            self.tgt_texts = tgt_texts

        # Prepare tokenized texts here to have them ready in __getitem__ method
        # This will speed up the batch preparation in the dataloader but it will raise system memory usage since the tokenized sequences are cached as dataset attribute
        # (The masks are also cached but their impact is negligible since they are arrays of boolean values)
        print("Caching encoded sequences in memory, this may take some time...")
        self.src_encoded_sequences, self.tgt_encoded_sequences = [], []
        self.src_masks, self.tgt_masks = [], []
        for src_text, tgt_text in zip(src_texts, tgt_texts, strict=False):
            src_text_tokenized = src_tokenizer.tokenize(src_text)
            tgt_text_tokenized = tgt_tokenizer.tokenize(tgt_text)

            # Exclude sequences that exceed max_sequence_length
            if max_sequence_length is not None and max_sequence_length > 0:
                if (
                    len(src_text_tokenized) > max_sequence_length - 2
                ):  # -2 accounts for SOS and EOS tokens that will be added in the __getitem__ method
                    continue
                if len(tgt_text_tokenized) > max_sequence_length - 2:
                    continue
                self.src_texts.append(src_text)
                self.tgt_texts.append(tgt_text)

            # src and tgt sequence lengths must be the same to properly compute cross attention
            # The smaller sequence will be padded to the length of the longer sequence
            # Attention mask ensure no attention is computed with pad tokens
            max_seq_len = max(len(src_text_tokenized), len(tgt_text_tokenized))

            # Tokenize texts
            src_tokens, src_mask = self.src_tokenizer.encode(src_text_tokenized, pad_to_len=max_seq_len)
            tgt_tokens, tgt_mask = self.tgt_tokenizer.encode(tgt_text_tokenized, pad_to_len=max_seq_len)

            self.src_encoded_sequences.append(src_tokens)
            self.tgt_encoded_sequences.append(tgt_tokens)
            self.src_masks.append(src_mask)
            self.tgt_masks.append(tgt_mask)

    def _build_vocabs(self, vocab_min_freq: int = 2, extend_with_glove: bool = False, **kwargs) -> None:
        """Build vocabularies for tokenizers."""

        print("Building vocabs, it may take a few minutes...")

        # Provides lists of tokens. Here the lists are not converted to sets cause the tokenizer may need the token frequencies
        src_tokens = [token for text in self.src_texts for token in self.src_tokenizer.tokenize(text)]
        tgt_tokens = [token for text in self.tgt_texts for token in self.tgt_tokenizer.tokenize(text)]

        self.src_tokenizer.build_vocab(
            src_tokens,
            min_freq=vocab_min_freq,
            extend_with_glove=bool(
                extend_with_glove and self.src_lang == "en"
            ),  # GloVe is trained on english only datasets so it doesn't make sense to extend non english vocabs
            glove_version=kwargs.get("glove_version", "glove.2024.wikigiga.50d"),
        )
        self.tgt_tokenizer.build_vocab(
            tgt_tokens,
            min_freq=vocab_min_freq,
            extend_with_glove=bool(extend_with_glove and self.tgt_lang == "en"),
            glove_version=kwargs.get("glove_version", "glove.2024.wikigiga.50d"),
        )

    def __len__(self):
        return len(self.src_texts)

    def __getitem__(self, index: int) -> dict[str, torch.Tensor | str]:
        src_tokens = self.src_encoded_sequences[index]
        tgt_tokens = self.tgt_encoded_sequences[index]
        src_mask = self.src_masks[index]
        tgt_mask = self.tgt_masks[index]
        src_text = self.src_texts[index]
        tgt_text = self.tgt_texts[index]

        return {
            "src": torch.tensor(src_tokens, dtype=torch.long),
            "tgt": torch.tensor(tgt_tokens, dtype=torch.long),
            "src_mask": torch.tensor(src_mask, dtype=torch.bool),
            "tgt_mask": torch.tensor(tgt_mask, dtype=torch.bool),
            "src_text": src_text,
            "tgt_text": tgt_text,
        }


def batch_collate_fn(
    batch: list[dict[str, torch.Tensor | str]],
    src_pad_token_id: int,
    tgt_pad_token_id: int,
    pad_all_to_len: int = -1,
) -> dict[str, torch.Tensor | list[str]]:
    """Used to tell the Dataloader how to properly build a batch.

    In order to correctly build a batch every sequence in it has to have the same length,
    so it pads the small sequences to the longest one. It does it for `src`, `tgt`, `src_mask` and `tgt_mask`.

    Note:
        This function needs two `PAD` token ids since in this Trasformer implementation there are 2 distinct tokenizers
        with their own vocabulary.
        Each vocabulary is built independently and in parallel, so there's no guarantee that it will have the same ID in both.

    Tip:
        By padding all sequences in the dataset to the same length higher GPU usage can achieved.

        It has to be coupled with `torch.compile` usage and with the dynamic cudagraph compilation disabled (`torch._inductor.config.triton.cudagraph_skip_dynamic_graphs = False`)

    Args:
        batch (dict[str, torch.Tensor  |  list[str]]): Batch of token ids and masks.
        src_pad_token_id (int): Pad token for source sequences.
        tgt_pad_token_id (int): Pad token for target sequences.
        pad_all_to_len (int): Sequence length to pad all sequences. If -1 it gets ignored. Defaults to -1.

    Returns:
        dict[str, torch.Tensor | list[str]]: Batch with padded sequences.
    """

    keys = batch[0].keys()
    result = {}

    for key in keys:
        field_data = [sample[key] for sample in batch]

        if isinstance(field_data[0], torch.Tensor):
            if key == "src":
                pad_token_id = src_pad_token_id
            elif key == "tgt":
                pad_token_id = tgt_pad_token_id
            else:  # mask tensors
                pad_token_id = 0  # will be replaced with False

            if pad_all_to_len != -1:
                num_padding_values = pad_all_to_len - field_data[0].shape[0]
                if num_padding_values >= 0:
                    field_data[0] = F.pad(field_data[0], (0, num_padding_values), value=pad_token_id)

            padded_seq = pad_sequence(
                typing.cast(list[torch.Tensor], field_data), batch_first=True, padding_value=pad_token_id
            )
            result[key] = padded_seq
        else:  # src_text and tgt_text
            result[key] = field_data

    return result


def build_data_utils(
    config: DictConfig | ListConfig, return_all: bool = False, **kwargs
) -> (
    tuple[DataLoader, DataLoader]
    | tuple[DataLoader, DataLoader, TranslationDataset, TranslationDataset, BaseTokenizer, BaseTokenizer]
):
    """Build tokenizers, datasets and dataloaders for Machine Translation.

    Args:
        config (DictConfig | ListConfig): Configuration object from omegaconf.
        return_all (bool, optional): Whether to return dataloaders, datasets and tokenizers. Defaults to False.

    Returns:
        tuple[DataLoader, DataLoader] | tuple[DataLoader, DataLoader, TranslationDataset, TranslationDataset, BaseTokenizer, BaseTokenizer]: Dataloaders or dataloaders, datasets and tokenizers.
    """

    data = load_dataset(config.dataset.dataset_id, config.dataset.dataset_name, cache_dir=config.cache_ds_path)["train"]

    # TODO If resuming tokenizers from pretrained add check to ensure src and tgt languages matched between the provided config and the checkpoint one
    src_lang = config.dataset.src_lang
    tgt_lang = config.dataset.tgt_lang

    # Downsample the dataset. Mainly for computational contraints and to make tests.
    if config.dataset.max_len != -1:
        data = data.select(range(config.dataset.max_len))

    print(f"Train test splitting - {config.dataset.train_split}/{1 - float(config.dataset.train_split):.2f}")
    split = data.train_test_split(train_size=config.dataset.train_split, seed=config.seed, shuffle=False)
    train_data = split["train"]
    test_data = split["test"]

    try:
        train_src_texts, train_tgt_texts = [], []
        for text in train_data["translation"]:
            train_src_texts.append(text[src_lang])
            train_tgt_texts.append(text[tgt_lang])
        test_src_texts, test_tgt_texts = [], []
        for text in test_data["translation"]:
            test_src_texts.append(text[src_lang])
            test_tgt_texts.append(text[tgt_lang])
    except KeyError as err:  # Mainly here to ensure consistency when resuming model checkpoint
        raise DatasetLanguagesNotAvailableError(src_lang, tgt_lang) from err

    # Build tokenizers and vocabs. Both src and tgt tokenizers vocabs are built using the training data
    special_tokens = {
        "sos_token": config.tokenizer.sos_token,
        "eos_token": config.tokenizer.eos_token,
        "pad_token": config.tokenizer.pad_token,
        "unk_token": config.tokenizer.unk_token,
    }

    # Get tokenizers from pretrained
    if "src_tokenizer" in kwargs and "tgt_tokenizer" in kwargs:
        print("Getting tokenizers from pretrained...")
        src_tokenizer, tgt_tokenizer = kwargs.get("src_tokenizer"), kwargs.get("tgt_tokenizer")

    else:
        print("Building tokenizers and vocabularies...")
        tok_types_dict: dict[str, typing.Any] = {"word": WordTokenizer, "bpe": BPETokenizer}
        src_tokenizer = tok_types_dict[config.tokenizer.type](
            special_tokens,
            tokenizer_max_len=config.tokenizer.max_seq_len,
            max_vocab_size=config.tokenizer.max_vocab_size,
        )
        tgt_tokenizer = tok_types_dict[config.tokenizer.type](
            special_tokens,
            tokenizer_max_len=config.tokenizer.max_seq_len,
            max_vocab_size=config.tokenizer.max_vocab_size,
        )

        # To build the vocabularies texts are not filtered based on tokenizer.max_sequence_length
        src_tokens = [token for text in train_src_texts for token in src_tokenizer.tokenize(text)]
        tgt_tokens = [token for text in train_tgt_texts for token in tgt_tokenizer.tokenize(text)]

        # BPE Tokenizer
        if config.tokenizer.type == "bpe":
            # TODO BPE should take all available text in order to extract statistics
            # and the tokenize method can be done only if the vocabulary is built
            src_tokenizer.build_vocab(src_tokens)
            tgt_tokenizer.build_vocab(tgt_tokens)
        else:
            src_tokenizer.build_vocab(
                src_tokens,
                min_freq=config.tokenizer.vocab_min_freq,
                extend_with_glove=bool(
                    src_lang == "en" and config.model_configs.pretrained_word_embeddings == "GloVe"
                ),  # GloVe is trained on english only datasets so it doesn't make sense to extend non english vocabs
                glove_version=config.model_configs[config.chosen_model_size].glove_version,
            )
            tgt_tokenizer.build_vocab(
                tgt_tokens,
                min_freq=config.tokenizer.vocab_min_freq,
                extend_with_glove=bool(tgt_lang == "en" and config.model_configs.pretrained_word_embeddings == "GloVe"),
                glove_version=config.model_configs[config.chosen_model_size].glove_version,
            )

    config.tokenizer.src_sos_token_idx = src_tokenizer.sos_token_idx
    config.tokenizer.src_eos_token_idx = src_tokenizer.eos_token_idx
    config.tokenizer.src_pad_token_idx = src_tokenizer.pad_token_idx
    config.tokenizer.src_unk_token_idx = src_tokenizer.unk_token_idx
    config.tokenizer.tgt_sos_token_idx = tgt_tokenizer.sos_token_idx
    config.tokenizer.tgt_eos_token_idx = tgt_tokenizer.eos_token_idx
    config.tokenizer.tgt_pad_token_idx = tgt_tokenizer.pad_token_idx
    config.tokenizer.tgt_unk_token_idx = tgt_tokenizer.unk_token_idx

    print("Building datasets...")
    train_dataset = TranslationDataset(
        train_src_texts,
        train_tgt_texts,
        src_tokenizer,
        tgt_tokenizer,
        src_lang=config.dataset.src_lang,
        tgt_lang=config.dataset.tgt_lang,
        max_sequence_length=config.tokenizer.max_seq_len,
    )
    test_dataset = TranslationDataset(
        test_src_texts,
        test_tgt_texts,
        src_tokenizer,
        tgt_tokenizer,
        src_lang=config.dataset.src_lang,
        tgt_lang=config.dataset.tgt_lang,
        max_sequence_length=config.tokenizer.max_seq_len,
    )

    print(f"Train dataset length: {len(train_dataset)}")
    print(f"Test dataset length: {len(test_dataset)}")

    train_dataloader = DataLoader(
        train_dataset,
        batch_size=config.train_dataloader.batch_size,
        num_workers=config.train_dataloader.num_workers,
        collate_fn=partial(
            batch_collate_fn,
            src_pad_token_id=config.tokenizer.src_pad_token_idx,
            tgt_pad_token_id=config.tokenizer.tgt_pad_token_idx,
            pad_all_to_len=config.tokenizer.max_seq_len if config.train_dataloader.pad_all_to_max_len else -1,
        ),
        shuffle=config.train_dataloader.shuffle,
        drop_last=config.train_dataloader.drop_last,
        persistent_workers=True,  # If True, the dataloader will not shut down the worker processes after a dataset has been consumed once. This allows to maintain the workers Dataset instances alive
        pin_memory=torch.cuda.is_available(),  # Pinned memory is not used anyway by torch where no accelerator is employed.
        prefetch_factor=config.train_dataloader.prefetch_factor,
    )
    test_dataloader = DataLoader(
        test_dataset,
        batch_size=config.test_dataloader.batch_size,
        num_workers=config.test_dataloader.num_workers,
        collate_fn=partial(
            batch_collate_fn,
            src_pad_token_id=config.tokenizer.src_pad_token_idx,
            tgt_pad_token_id=config.tokenizer.tgt_pad_token_idx,
            pad_all_to_len=config.tokenizer.max_seq_len if config.test_dataloader.pad_all_to_max_len else -1,
        ),
        shuffle=config.test_dataloader.shuffle,
        drop_last=config.test_dataloader.drop_last,
        persistent_workers=True,
        pin_memory=torch.cuda.is_available(),
        prefetch_factor=config.train_dataloader.prefetch_factor,
    )

    print(f"Train dataloader length: {len(train_dataloader)}")
    print(f"Test dataloader length: {len(test_dataloader)}")

    if return_all:
        return train_dataloader, test_dataloader, train_dataset, test_dataset, src_tokenizer, tgt_tokenizer
    return train_dataloader, test_dataloader



# Transformer inference script
#
# Author: Giovanni Spadaro - https://giovannispadaro.it
# Project: https://github.com/Giovo17/tfs-mt
# Documentation: https://giovo17.github.io/tfs-mt
#
# Copyright (c) Giovanni Spadaro.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in the root directory of this source tree.

import torch

from tfs_mt.architecture import build_model
from tfs_mt.data_utils import WordTokenizer
from tfs_mt.decoding_utils import greedy_decoding

base_url = "https://huggingface.co/giovo17/tfs-mt/resolve/main/"
src_tokenizer = WordTokenizer.from_pretrained(base_url + "src_tokenizer_word.json")
tgt_tokenizer = WordTokenizer.from_pretrained(base_url + "tgt_tokenizer_word.json")

model = build_model(
    config="https://huggingface.co/giovo17/tfs-mt/resolve/main/config-lock.yaml",
    from_pretrained=True,
    model_path="https://huggingface.co/giovo17/tfs-mt/resolve/main/model.safetensors",
)

device = "cuda:0" if torch.cuda.is_available() else "cpu"
model.to(device)
model.eval()

input_tokens, input_mask = src_tokenizer.encode("Hi, how are you?")

output = greedy_decoding(model, tgt_tokenizer, input_tokens, input_mask)[0]
print(output)



# Transformer train script
#
# Author: Giovanni Spadaro - https://giovannispadaro.it
# Project: https://github.com/Giovo17/tfs-mt
# Documentation: https://giovo17.github.io/tfs-mt
#
# Copyright (c) Giovanni Spadaro.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in the root directory of this source tree.

import argparse
import gc
import logging
import os
import typing
from datetime import datetime
from functools import partial
from pprint import pformat

import torch
from dotenv import load_dotenv
from ignite.engine import Events
from ignite.metrics import Bleu, Loss, Rouge
from ignite.utils import manual_seed, setup_logger
from omegaconf import OmegaConf
from torch.nn import CrossEntropyLoss
from torch.optim import AdamW
from torch.optim.lr_scheduler import LambdaLR
from torch.profiler import ProfilerActivity
from torchinfo import summary

from tfs_mt.architecture import build_model
from tfs_mt.data_utils import build_data_utils
from tfs_mt.decoding_utils import greedy_decoding
from tfs_mt.ignite_custom_utils.metrics import Perplexity
from tfs_mt.training_utils import (
    KLDivLabelSmoothingLoss,
    get_device,
    get_param_groups,
    log_metrics,
    loss_metric_transform,
    nlp_metric_transform,
    resume_from_ckpt,
    s3_upload,
    save_config,
    setup_early_stopping,
    setup_evaluator,
    setup_exp_logging,
    setup_handlers,
    setup_lr_lambda_fn,
    setup_output_dir,
    setup_trainer,
)

load_dotenv()
# os.environ["PYTORCH_CUDA_ALLOC_CONF"] = "expandable_segments:True"
# os.environ["TORCHINDUCTOR_DISABLE_CUDAGRAPH"] = "1"
os.environ["TOKENIZERS_PARALLELISM"] = "false"  # remove tokenizer parallelism warning


# Torch profiler activities
activities = [ProfilerActivity.CPU]

if torch.cuda.is_available():
    # Consider if the batches aren't aligned with the same sequence length.
    # If so, when compiling the model CUDA creates separate cudagraphs for each different input shape.
    # Set to True to avoid this overhead when having batched of different input lengths.
    # Set to False to create cudagraphs and get speedup, but ensure the batches have all the same sequence length to avoid cudagraph recomputation overhead.
    # NOTE The best solution may be in the middle:
    #   First of all extract K different sequence lengths such that a good samples clusterization is achieved.
    #   (good is referred in terms of how similar, in terms of how long they are, sequences inside the cluster and how dissimilar the clusters are)
    #   When collating batches the samples inside have to be in the same previously assigned cluster and the batch-assigned padding sequence length has to be the cluster max sequence length.
    #   This way the number of different sequence length reduces drastically and few cudagraphs can be created (so the following option has to be set to False).
    #   This approach is demanded to future development.
    # Current approach:
    #   Set tokenizer max sequence length to the max sequence length encountered in training data
    #   Pad to tokenizer max sequence length
    #   Compile unique cudagraph
    torch._inductor.config.triton.cudagraph_skip_dynamic_graphs = False

    # Allow matmul with FP32 precision to leverage NVIDIA tensor cores. More here: https://docs.pytorch.org/docs/main/notes/cuda.html#tensorfloat-32-tf32-on-ampere-and-later-devices
    torch.backends.fp32_precision = "ieee"
    torch.backends.cuda.matmul.fp32_precision = "ieee"
    torch.backends.cudnn.fp32_precision = "ieee"
    torch.backends.cudnn.benchmark = True
    activities += [ProfilerActivity.CUDA]


wandb_api_key = os.getenv("WANDB_API_KEY")
os.environ["WANDB_API_KEY"] = wandb_api_key


class TooManyWarmupItersError(Exception):
    def __init__(self, warmup_iters, total_iters):
        msg = f"The number of warmup iterations cannot be greater than 50% the total number of iterations, \
        got warmup_iters: {warmup_iters}, total_iterations: {total_iters}"
        super().__init__(msg)


def run(config, enable_log_ckpt=True):
    manual_seed(config.seed)
    output_dir = setup_output_dir(config)
    config.output_dir = output_dir
    save_config(config, config.output_dir, enable_ckpt=enable_log_ckpt)

    # Setup logger
    logger = setup_logger(level=logging.INFO, filepath=os.path.join(config.output_dir, "training-info.log"))
    logger.info("Configuration: \n%s", pformat(config))

    # Resume tokenizer from pretrained and build data utils
    if config.ckpt_path_to_resume_from is not None or config.tokenizers_resume_path is not None:
        src_tokenizer, tgt_tokenizer = typing.cast(
            tuple[typing.Any, typing.Any],
            resume_from_ckpt(
                checkpoint_path=config.ckpt_path_to_resume_from
                if config.ckpt_path_to_resume_from is not None
                else config.tokenizers_resume_path,
                logger=logger,
                resume_tokenizers=True,
                tokenizers_type=config.tokenizer.type,
            ),
        )
        train_dataloader, test_dataloader = typing.cast(
            tuple[typing.Any, typing.Any],
            build_data_utils(config, src_tokenizer=src_tokenizer, tgt_tokenizer=tgt_tokenizer),
        )

    else:  # Build data utils from scratch
        train_dataloader, test_dataloader, _, _, src_tokenizer, tgt_tokenizer = typing.cast(
            tuple[typing.Any, typing.Any, typing.Any, typing.Any, typing.Any, typing.Any],
            build_data_utils(config, return_all=True),
        )

    # Save tokenizers
    src_tokenizer.to_json(config.output_dir + f"/src_tokenizer_{config.tokenizer.type}.json")
    tgt_tokenizer.to_json(config.output_dir + f"/tgt_tokenizer_{config.tokenizer.type}.json")
    if config.s3_bucket_name is not None and enable_log_ckpt:
        s3_upload(
            filepath=config.output_dir + f"/src_tokenizer_{config.tokenizer.type}.json",
            bucket=config.s3_bucket_name,
            s3_key=f"{config.model_name}/src_tokenizer_{config.tokenizer.type}.json",
        )
        s3_upload(
            filepath=config.output_dir + f"/tgt_tokenizer_{config.tokenizer.type}.json",
            bucket=config.s3_bucket_name,
            s3_key=f"{config.model_name}/tgt_tokenizer_{config.tokenizer.type}.json",
        )
        logger.info(f"Uploaded tokenizers to s3://{config.s3_bucket_name}")

    config.src_tokenizer_vocab_size = src_tokenizer.vocab_size
    config.tgt_tokenizer_vocab_size = tgt_tokenizer.vocab_size
    logger.info(f"Vocabulary size of source tokenizer: {config.src_tokenizer_vocab_size}")
    logger.info(f"Vocabulary size of target tokenizer: {config.tgt_tokenizer_vocab_size}")

    config.num_train_iters_per_epoch = len(train_dataloader)
    config.num_test_iters_per_epoch = len(test_dataloader)
    logger.info(f"Number of train iterations per epoch: {config.num_train_iters_per_epoch}")
    logger.info(f"Number of test iterations per epoch: {config.num_test_iters_per_epoch}")

    # Raise exception if warmup iterations are more than 50% of total iterations
    if (
        config.training_hp.lr_scheduler.warmup_iters
        > 0.5 * config.training_hp.num_epochs * config.num_train_iters_per_epoch
    ):
        raise TooManyWarmupItersError(
            config.training_hp.lr_scheduler.warmup_iters,
            config.training_hp.num_epochs * config.num_train_iters_per_epoch,
        )

    # Initialize model, optimizer, loss function, device or resume from checkpoint
    device = get_device()
    logger.info(f"Using device: {device}")

    model = build_model(config, src_tokenizer, tgt_tokenizer)
    model.to(device)
    if torch.cuda.is_available():
        model = torch.compile(model, mode=config.training_hp.torch_compile_mode)

    logger.info(
        summary(
            typing.cast(torch.nn.Module, model),
            [(16, 128), (16, 128), (16, 128), (16, 128)],
            dtypes=[torch.long, torch.long, torch.bool, torch.bool],
        )
    )
    logger.info(f"Total number of parameters: {sum(p.numel() for p in model.parameters()) / 1e6:.2f} M")

    model_param_groups = get_param_groups(model.named_parameters(), config.training_hp.optimizer.weight_decay)

    optim_type = {"AdamW": AdamW}
    optimizer = optim_type[config.training_hp.optimizer.type](
        model_param_groups,
        lr=1,  # This way the learning rate values will be entirely managed by the learning rate scheduler
        betas=(config.training_hp.optimizer.beta1, config.training_hp.optimizer.beta2),
    )

    if config.training_hp.loss.type == "crossentropy":
        loss_fn = CrossEntropyLoss(
            # Ignore padding tokens in loss computation
            ignore_index=tgt_tokenizer.pad_token_idx,
            # This will average the loss over batch_size * sequence_length (pad tokens don't contribute to sequence_length)
            reduction="sum",  # The loss will be manually averaged by the number of processed tokens
            # During training, we employed label smoothing of value 0.1 (Attention is all you need page 8)
            # This reduces overconfidence and improves generalization
            label_smoothing=config.training_hp.loss.label_smoothing,
        ).to(device=device)

    elif config.training_hp.loss.type == "KLdiv-labelsmoothing":
        loss_fn = KLDivLabelSmoothingLoss(
            vocab_size=tgt_tokenizer.vocab_size,
            padding_idx=tgt_tokenizer.pad_token_idx,
            smoothing=config.training_hp.loss.label_smoothing,
        ).to(device=device)

    else:
        raise ValueError(f"Invalid loss type, got {config.training_hp.loss.type}")

    # Initialize learning rate scheduler
    lr_lambda = setup_lr_lambda_fn(config)
    lr_scheduler = LambdaLR(optimizer, lr_lambda=lr_lambda)

    # Setup metrics
    metrics = {
        "Bleu": Bleu(
            ngram=4, smooth="smooth1", output_transform=partial(nlp_metric_transform, tgt_tokenizer=tgt_tokenizer)
        ),
        "Bleu_smooth_2": Bleu(
            ngram=4, smooth="smooth2", output_transform=partial(nlp_metric_transform, tgt_tokenizer=tgt_tokenizer)
        ),
        "Rouge": Rouge(
            variants=["L", 2],
            multiref="best",
            output_transform=partial(nlp_metric_transform, tgt_tokenizer=tgt_tokenizer),
        ),
        "Perplexity": Perplexity(
            criterion=loss_fn,
            pad_idx=tgt_tokenizer.pad_token_idx,
            output_transform=partial(loss_metric_transform, loss_type=config.training_hp.loss.type),
            device=device,
        ),
        "Loss": Loss(loss_fn, output_transform=partial(loss_metric_transform, loss_type=config.training_hp.loss.type)),
    }

    # Setup trainer and evaluator
    trainer = setup_trainer(config, typing.cast(torch.nn.Module, model), optimizer, lr_scheduler, loss_fn, device)
    test_evaluator = setup_evaluator(config, model, metrics, device)

    # Setup engines logger with python logging print training configurations
    trainer.logger = logger
    test_evaluator.logger = logger

    # Setup ignite handlers
    to_save_train = {
        "model": model,
        "optimizer": optimizer,
        "trainer": trainer,
        "lr_scheduler": lr_scheduler,
    }
    to_save_test = {"model": model}
    # When enable_log_ckpt is False the two returned ckpt handlers will be None
    setup_handlers(trainer, test_evaluator, config, to_save_train, to_save_test, enable_ckpt=enable_log_ckpt)

    # Time profiling
    # profiler = HandlersTimeProfiler()
    # profiler.attach(trainer)

    # Experiment tracking
    if enable_log_ckpt:
        exp_wandb_logger, exp_trackio_logger = setup_exp_logging(
            config,
            trainer,
            optimizer,
            evaluator=test_evaluator,
            metrics=metrics,
            model=model,
            return_all_loggers=True,
        )

    # Print metrics to the stderr with "add_event_handler" method for training stats
    # More on ignite Events: https://docs.pytorch.org/ignite/generated/ignite.engine.events.Events.html
    trainer.add_event_handler(
        Events.ITERATION_COMPLETED(every=config.log_every_iters),
        log_metrics,
        tag="train",
    )

    # Clean GPU cache
    @test_evaluator.on(Events.EPOCH_STARTED | Events.EPOCH_COMPLETED)
    def cleanup_memory(engine):
        gc.collect()
        if torch.cuda.is_available():
            torch.cuda.empty_cache()
            torch.cuda.reset_peak_memory_stats()

    # Run evaluators at every training epoch end and print metrics to the stderr
    @trainer.on(Events.EPOCH_COMPLETED)
    def run_test_eval():
        test_evaluator.run(test_dataloader)
        log_metrics(test_evaluator, "test")

    # Run evaluator when trainer starts to make sure it works.
    @trainer.on(Events.STARTED)
    def run_test_eval_on_start():
        test_evaluator.run(test_dataloader)
        # Attach early stopping to test_evaluator here cause it relies on evaluator metrics dict to be defined and it is non-empty only after first run
        if config.training_hp.early_stopping.enabled:
            logger.info("Enabled early stopping.")
            setup_early_stopping(trainer, test_evaluator, config)

    # Decode a sequence to debug training
    @test_evaluator.on(Events.EPOCH_COMPLETED(every=1))
    def run_decoding_debug():
        nr_sequences = 3
        sample_batch = next(test_dataloader.__iter__())
        decoded_seq_batch = greedy_decoding(
            model=model,
            tgt_tokenizer=tgt_tokenizer,
            src_tokens=sample_batch["src"][:nr_sequences],
            src_mask=sample_batch["src_mask"][:nr_sequences],
            max_target_tokens=config.tokenizer.max_seq_len,
            output_mode="str",
        )
        test_evaluator.logger.info(f"Source sequences: \n{sample_batch['src_text'][:nr_sequences]}")
        test_evaluator.logger.info(f"Decoded sequences: \n{decoded_seq_batch}")

    # Log time profiling
    # @trainer.on(Events.EPOCH_COMPLETED)
    # def log_intermediate_results():
    #    profiler.print_results(profiler.get_results())

    # Resume from checkpoint if option available in config
    if config.ckpt_path_to_resume_from is not None:
        resume_from_ckpt(
            config.ckpt_path_to_resume_from,
            to_load=to_save_train,
            device=device,
            logger=logger,
            strict=True,
        )

    # Save config-lock
    save_config(config, output_dir, enable_ckpt=enable_log_ckpt)

    # Setting up the profiler scheduler
    # https://docs.pytorch.org/tutorials/recipes/recipes/profiler_recipe.html#using-profiler-to-analyze-long-running-jobs
    prof_schedule = torch.profiler.schedule(wait=1, warmup=1, active=1, repeat=1)  # It will record iterations 2

    def trace_handler(p):
        output = p.key_averages().table(sort_by="self_cuda_time_total", row_limit=10)
        logger.info(f"Profiler results:\n{output}")
        p.export_chrome_trace(os.path.join(config.output_dir, "trace.json"))
        if config.s3_bucket_name is not None and enable_log_ckpt:
            s3_upload(
                filepath=os.path.join(config.output_dir, "trace.json"),
                bucket=config.s3_bucket_name,
                s3_key=f"{config.model_name}/trace.json",
            )
            logger.info("Uploaded trace to s3")

    # Run and profile training
    # PyTorch official tutorial on profiling: https://docs.pytorch.org/tutorials/recipes/recipes/profiler_recipe.html
    with torch.profiler.profile(
        activities=activities,
        schedule=prof_schedule,
        on_trace_ready=trace_handler,
        record_shapes=True,
        profile_memory=True,
        with_stack=True,
    ) as prof:
        trainer.add_event_handler(Events.ITERATION_COMPLETED, lambda _: prof.step())
        trainer.run(train_dataloader, max_epochs=config.training_hp.num_epochs)

    # Close loggers and upload local log file to s3 if configured
    if enable_log_ckpt:
        exp_wandb_logger.close()
        exp_trackio_logger.close()
        # profiler.write_results(config.output_dir + "/time_profiling.csv")
    if config.s3_bucket_name is not None and enable_log_ckpt:
        s3_upload(
            filepath=os.path.join(config.output_dir, "training-info.log"),
            bucket=config.s3_bucket_name,
            s3_key=f"{config.model_name}/training-info.log",
        )


def parse_args():
    parser = argparse.ArgumentParser(description="Transformer training arguments")
    parser.add_argument(
        "-p",
        "--path",
        type=str,
        default="/".join(os.path.abspath(__file__).split("/")[:-2]),
        help="Path to the project folder (default: this script's directory).",
    )
    parser.add_argument(
        "-e",
        "--exec-mode",
        choices=["dev", "dummy"],
        default="dev",
        help="""
        Execution mode: 'dev' or 'dummy' (default: dev).
        Dev: train the model in developer way, all training features are enabled.
        Dummy: debug model training, this disables experiment tracking and checkpointing.
        """,
    )
    parser.add_argument(
        "-s",
        "--size",
        choices=["nano", "small", "base", "original"],
        default="nano",
        help="""
        Model size: 'nano', 'small' or 'base' (default: nano).
        Refer to configs/config.yml file for more info on sizes.
        """,
    )
    parser.add_argument(
        "-tl",
        "--time-limit",
        type=int,
        default=-1,
        help="Execution time limit defined in seconds (default: -1).",
    )
    parser.add_argument(
        "-mts",
        "--max-train-samples",
        type=int,
        default=-1,
        help="Maximum number of training samples to use. -1 means use all available samples (default: -1).",
    )
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()

    base_path = args.path if os.path.isdir(args.path) else "/".join(os.path.abspath(__file__).split("/")[:-2])
    config_path = os.path.join(base_path, "src/tfs_mt/configs/config.yml")

    config = OmegaConf.load(config_path)

    config.base_path = base_path
    config.output_dir = os.path.join(base_path, "data/output")
    config.cache_ds_path = os.path.join(base_path, "data")
    config.chosen_model_size = args.size
    config.time_limit_sec = args.time_limit if args.time_limit > 0 else -1
    config.wandb_organization = os.getenv("WANDB_ORGANIZATION")

    # Apply max_train_samples if specified
    if args.max_train_samples > 0:
        config.dataset.max_len = args.max_train_samples

    # Detect Kaggle enviroment and make adjustments
    if os.path.exists("/kaggle"):
        print("Kaggle environment detected. Overriding some config options...")
        config.time_limit_sec = (
            args.time_limit if (args.time_limit <= 11.5 * 3600 and args.time_limit > 0) else 11.5 * 3600
        )
        config.training_hp.amp_dtype = "float16"  # Neither the T4 nor the P100 support bfloat16
        config.train_dataloader.num_workers = 4
        config.test_dataloader.num_workers = 4
        config.training_hp.torch_compile_mode = None

    # Setup dummy training. Mainly used for debugging
    print(f"Execution mode: {args.exec_mode}")
    if args.exec_mode == "dummy":
        print("Experiment logging and checkpointing are disabled!")
        config.dataset.max_len = 1_000
        config.training_hp.lr_scheduler.warmup_iters = (
            5  # if it's too high it raises TooManyWarmupItersError during training setup
        )
        config.log_every_iters = 1
        config.training_hp.num_epochs = 2
        config.training_hp.torch_compile_mode = None

    # Fix batch sizes if they are larger than train/test data (dataset_max_len * split_proportion)
    if config.dataset.max_len > 0 and config.train_dataloader.batch_size > int(
        config.dataset.max_len * config.dataset.train_split
    ):
        config.train_dataloader.batch_size = int(config.dataset.max_len * config.dataset.train_split)
    if config.dataset.max_len > 0 and config.test_dataloader.batch_size > int(
        config.dataset.max_len * (1 - config.dataset.train_split)
    ):
        config.test_dataloader.batch_size = int(config.dataset.max_len * (1 - config.dataset.train_split))

    config.model_name = f"{config.model_base_name}_{config.chosen_model_size}_{datetime.now().strftime('%y%m%d-%H%M')}"

    config.exec_mode = args.exec_mode

    if args.exec_mode == "dummy":
        run(config, enable_log_ckpt=False)
    elif args.exec_mode == "dev":
        run(config, enable_log_ckpt=True)



# Checkpoint uploader script
#
# Author: Giovanni Spadaro - https://giovannispadaro.it
# Project: https://github.com/Giovo17/tfs-mt
# Documentation: https://giovo17.github.io/tfs-mt
#
# Copyright (c) Giovanni Spadaro.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in the root directory of this source tree.

import argparse
import os
import tempfile
from typing import Any

import torch
from dotenv import load_dotenv
from huggingface_hub import HfApi, create_repo
from huggingface_hub.errors import RepositoryNotFoundError
from omegaconf import DictConfig, OmegaConf
from safetensors.torch import save_file

from tfs_mt.architecture import Transformer
from tfs_mt.ignite_custom_utils.checkpoint import S3Saver

load_dotenv(override=True)


class CheckpointUploadError(Exception):
    """Base exception for checkpoint upload errors."""

    pass


class S3DownloadError(CheckpointUploadError):
    """Exception raised when S3 download fails."""

    def __init__(self, path: str, original_error: Exception):
        self.path = path
        self.original_error = original_error
        super().__init__(f"Failed to download from S3: {path}. Error: {original_error}")


def parse_s3_path(s3_path: str) -> tuple[str, str]:
    """Parse S3 path into bucket and key."""
    _, _, path = s3_path.partition("s3://")
    bucket, _, key = path.partition("/")

    if not bucket or not key:
        raise ValueError(f"Invalid S3 path: {s3_path}. Expected format: s3://bucket/key")

    return bucket, key


def download_from_s3(s3_client, bucket: str, key: str, local_path: str):
    """Download a file from S3."""
    try:
        s3_client.download_file(bucket, key, local_path)
    except Exception as e:
        raise S3DownloadError(f"s3://{bucket}/{key}", e) from e


def load_checkpoint_from_path(checkpoint_path: str, device: str = "cpu") -> dict[str, Any]:
    """
    Load checkpoint from local file or S3.

    Args:
        checkpoint_path: Path to checkpoint file (local or s3://)
        device: Device to load the checkpoint on

    Returns:
        Dictionary containing checkpoint data
    """
    is_s3 = checkpoint_path.startswith("s3://")

    if is_s3:
        bucket, key = parse_s3_path(checkpoint_path)
        s3 = S3Saver._make_s3_client()

        with tempfile.NamedTemporaryFile(suffix=".pt") as tmp:
            tmp_path = tmp.name

            print(f"Downloading checkpoint from S3: {checkpoint_path}")
            download_from_s3(s3, bucket, key, tmp_path)
            checkpoint = torch.load(tmp_path, map_location=device, weights_only=True)
            return checkpoint

    else:
        if not os.path.isfile(checkpoint_path):
            raise FileNotFoundError(f"Checkpoint not found: {checkpoint_path}")

        print(f"Loading checkpoint from local file: {checkpoint_path}")
        checkpoint = torch.load(checkpoint_path, map_location=device, weights_only=True)
        return checkpoint


def get_base_path(checkpoint_path: str) -> str:
    """Get the base directory path from checkpoint path."""
    if checkpoint_path.endswith((".pt", ".pth")):
        return "/".join(checkpoint_path.split("/")[:-1])
    return checkpoint_path


def load_config(config_path: str) -> DictConfig | None:
    """Load config-lock.yaml from local or S3 using OmegaConf."""
    is_s3 = config_path.startswith("s3://")

    if is_s3:
        bucket, key = parse_s3_path(config_path)
        s3 = S3Saver._make_s3_client()

        with tempfile.NamedTemporaryFile(suffix=".yaml", mode="w+") as tmp:
            tmp_path = tmp.name

            download_from_s3(s3, bucket, key, tmp_path)
            config = OmegaConf.load(tmp_path)
            return config

    else:
        if not os.path.isfile(config_path):
            print(f"Warning: Config file not found: {config_path}")
            return None

        config = OmegaConf.load(config_path)
        return config


def format_model_architecture(config: DictConfig, chosen_size: str) -> list[str]:
    """Format model architecture details for README."""
    lines = []

    if "model_configs" in config and chosen_size in config.model_configs:
        model_cfg = OmegaConf.to_container(config.model_configs[chosen_size], resolve=True)
        lines.append(f"**Model Size**: `{chosen_size}`")
        lines.append("")

        architecture_params = {
            "num_encoder_layers": "Encoder Layers",
            "num_decoder_layers": "Decoder Layers",
            "d_model": "Model Dimension",
            "num_heads": "Attention Heads",
            "d_ff": "FFN Dimension",
            "norm_type": "Normalization Type",
        }

        for key, label in architecture_params.items():
            if key in model_cfg:
                lines.append(f"- **{label}**: {model_cfg[key]}")

        if "dropout" in config.model_parameters:
            lines.append(f"- **Dropout**: {config.model_parameters.dropout}")

        # Embeddings info
        if "pretrained_word_embeddings" in config.model_configs:
            lines.append(f"- **Pretrained Embeddings**: {config.model_configs.pretrained_word_embeddings}")
        if "positional_embeddings" in config.model_configs:
            lines.append(f"- **Positional Embeddings**: {config.model_configs.positional_embeddings}")

        # GloVe details if available
        if "glove_version" in model_cfg:
            lines.append(f"- **GloVe Version**: {model_cfg['glove_version']}")

    return lines


def format_training_details(config: DictConfig) -> list[str]:
    """Format training hyperparameters for README."""
    lines = []

    hp = config.training_hp

    # Basic training settings
    lines.append(f"- **Epochs**: {hp.num_epochs}")
    amp_status = f"{hp.use_amp}"
    if hp.use_amp and "amp_dtype" in hp:
        amp_status += f" ({hp.amp_dtype})"
    lines.append(f"- **Mixed Precision**: {amp_status}")
    lines.append(f"- **Torch Compile Mode**: {hp.torch_compile_mode}")
    lines.append(f"- **Gradient Clipping**: {hp.max_gradient_norm}")

    # Loss configuration
    loss_cfg = OmegaConf.to_container(hp.loss, resolve=True)
    lines.append(f"- **Loss Type**: {loss_cfg.get('type', 'N/A')}")
    lines.append(f"- **Label Smoothing**: {loss_cfg['label_smoothing']}")

    # Optimizer
    opt_cfg = OmegaConf.to_container(hp.optimizer, resolve=True)
    lines.append(f"- **Optimizer**: {opt_cfg.get('type', 'N/A')}")
    lines.append(f"  - Weight Decay: {opt_cfg.get('weight_decay', 'N/A')}")
    lines.append(f"  - Beta1: {opt_cfg.get('beta1', 'N/A')}, Beta2: {opt_cfg.get('beta2', 'N/A')}")
    lines.append(f"  - Epsilon: {opt_cfg.get('eps', 'N/A')}")

    # Learning rate scheduler
    lr_cfg = OmegaConf.to_container(hp.lr_scheduler, resolve=True)
    lines.append(f"- **LR Scheduler**: {lr_cfg.get('type', 'N/A')}")
    lines.append(f"  - Min LR: {lr_cfg.get('min_lr', 'N/A')}")
    lines.append(f"  - Max LR: {lr_cfg.get('max_lr', 'N/A')}")
    lines.append(f"  - Warmup Iterations: {lr_cfg.get('warmup_iters', 'N/A')}")
    if "stable_iters_prop" in lr_cfg:
        lines.append(f"  - Stable Iterations Proportion: {lr_cfg['stable_iters_prop']}")

    return lines


def format_tokenizer_details(config: DictConfig) -> list[str]:
    """Format tokenizer configuration for README."""
    lines = []

    tok_cfg = OmegaConf.to_container(config.tokenizer, resolve=True)

    lines.append(f"- **Type**: {tok_cfg.get('type', 'N/A')}")
    lines.append(f"- **Max Sequence Length**: {tok_cfg.get('max_seq_len', 'N/A')}")
    lines.append(f"- **Max Vocabulary Size**: {tok_cfg.get('max_vocab_size', 'N/A')}")
    lines.append(f"- **Minimum Frequency**: {tok_cfg.get('vocab_min_freq', 'N/A')}")
    lines.append("")

    return lines


def format_dataset_details(config: DictConfig) -> list[str]:
    """Format dataset information for README."""
    lines = []

    ds_cfg = OmegaConf.to_container(config.dataset, resolve=True)

    lines.append(f"- **Task**: {ds_cfg.get('dataset_task', 'N/A')}")
    lines.append(f"- **Dataset ID**: `{ds_cfg.get('dataset_id', 'N/A')}`")
    lines.append(f"- **Dataset Name**: `{ds_cfg.get('dataset_name', 'N/A')}`")
    lines.append(f"- **Source Language**: {ds_cfg.get('src_lang', 'N/A')}")
    lines.append(f"- **Target Language**: {ds_cfg.get('tgt_lang', 'N/A')}")
    lines.append(f"- **Train Split**: {ds_cfg.get('train_split', 'N/A')}")

    return lines


def generate_readme(config: DictConfig, model_name: str) -> str:
    """Generate a comprehensive README for the Hugging Face repository."""

    # 1. Load template
    template_path = os.path.join(os.path.dirname(__file__), "tfs_mt/templates/model_README.md")
    if not os.path.exists(template_path):
        raise FileNotFoundError(f"README template not found at {template_path}")

    with open(template_path) as f:
        template = f.read()

    # 2. Prepare dynamic content
    src = config.dataset.get("src_lang", "")
    tgt = config.dataset.get("tgt_lang", "")

    datasets_section = ""
    if "dataset_id" in config.dataset:
        datasets_section = f"datasets:\n- {config.dataset.dataset_id}"

    inference_script_path = os.path.join(os.getcwd(), "src/inference.py")
    with open(inference_script_path) as f:
        inference_code_lines = f.readlines()

    inference_code_lines = [line.rstrip("\n") for line in inference_code_lines if not line.startswith("#")]
    if inference_code_lines and not inference_code_lines[0]:
        inference_code_lines.pop(0)
    inference_code = "\n".join(inference_code_lines)

    chosen_size = config.get("chosen_model_size", "unknown")
    model_architecture = "\n".join(format_model_architecture(config, chosen_size))
    tokenizer_details = "\n".join(format_tokenizer_details(config))
    dataset_details = "\n".join(format_dataset_details(config))
    full_config = OmegaConf.to_yaml(config, resolve=True).strip()

    # 3. Fill template
    replacements = {
        "{{src_lang}}": src,
        "{{tgt_lang}}": tgt,
        "{{datasets_section}}": datasets_section,
        "{{inference_code}}": inference_code,
        "{{model_architecture}}": model_architecture,
        "{{tokenizer_details}}": tokenizer_details,
        "{{dataset_details}}": dataset_details,
        "{{full_config}}": full_config,
    }

    readme = template
    for placeholder, value in replacements.items():
        readme = readme.replace(placeholder, str(value))

    return readme


def upload_to_huggingface(
    checkpoint_path: str,
    repo_id: str,
    hf_token: str | None = None,
    private: bool = True,
):
    """
    Upload model checkpoint, tokenizers, config, and README to Hugging Face.

    Args:
        checkpoint_path: Path to checkpoint (local or s3://)
        repo_id: Hugging Face repository ID (e.g., "username/model-name")
        hf_token: Hugging Face API token
        private: Whether to create a private repository
    """
    api = HfApi(token=hf_token)

    try:
        repo_files = api.list_repo_files(repo_id=repo_id)
    except RepositoryNotFoundError:
        repo_files = []

    # Create repository
    print(f"Creating/accessing repository: {repo_id}")
    try:
        create_repo(repo_id, private=private, token=hf_token, exist_ok=True)
    except Exception as e:
        print(f"Warning: Could not create repo (may already exist): {e}")

    # Get base path for tokenizers and config
    base_path = get_base_path(checkpoint_path)
    is_s3 = checkpoint_path.startswith("s3://")

    # Load checkpoint
    checkpoint = load_checkpoint_from_path(checkpoint_path, "cpu")

    # Save model to temporary file
    ckpt_dict = {
        "model": "model.pt",
        "optimizer": "optimizer.pt",
        "lr_scheduler": "lr_scheduler.pt",
        "trainer": "trainer.pt",
    }
    for key, filename in ckpt_dict.items():
        if key not in checkpoint:
            print(f"Skipping '{key}', not found in checkpoint.")
            continue

        with tempfile.NamedTemporaryFile(suffix=".pt") as tmp:
            tmp_path = tmp.name

            print(f"Saving {key} checkpoint...")
            torch.save(checkpoint[key], tmp_path)

            if filename not in repo_files:
                print(f"Uploading {key} checkpoint to repo as {filename}...")
                api.upload_file(
                    path_or_fileobj=tmp_path,
                    path_in_repo=filename,
                    repo_id=repo_id,
                    token=hf_token,
                )

    config_path = f"{base_path}/config-lock.yaml"
    config = load_config(config_path)

    # Upload tokenizers
    if is_s3:
        src_tokenizer_filename = f"src_tokenizer_{config.tokenizer.type}.json"
        tgt_tokenizer_filename = f"tgt_tokenizer_{config.tokenizer.type}.json"

        bucket, base_key = parse_s3_path(base_path)
        src_tokenizer_key = f"{base_key}/{src_tokenizer_filename}"
        tgt_tokenizer_key = f"{base_key}/{tgt_tokenizer_filename}"
        s3 = S3Saver._make_s3_client()

        with tempfile.TemporaryDirectory() as tmp_dir:
            try:
                src_tokenizer_tmp_path = os.path.join(tmp_dir, src_tokenizer_filename)
                tgt_tokenizer_tmp_path = os.path.join(tmp_dir, tgt_tokenizer_filename)

                print("Downloading tokenizers from S3...")
                download_from_s3(s3, bucket, src_tokenizer_key, src_tokenizer_tmp_path)
                download_from_s3(s3, bucket, tgt_tokenizer_key, tgt_tokenizer_tmp_path)

                if src_tokenizer_filename not in repo_files or tgt_tokenizer_filename not in repo_files:
                    print("Uploading tokenizers...")
                    api.upload_file(
                        path_or_fileobj=src_tokenizer_tmp_path,
                        path_in_repo=src_tokenizer_filename,
                        repo_id=repo_id,
                        token=hf_token,
                    )
                    api.upload_file(
                        path_or_fileobj=tgt_tokenizer_tmp_path,
                        path_in_repo=tgt_tokenizer_filename,
                        repo_id=repo_id,
                        token=hf_token,
                    )
            except S3DownloadError as e:
                print(f"Warning: Could not upload tokenizers: {e}")

    else:
        src_tokenizer_path = f"{base_path}/{src_tokenizer_filename}"
        tgt_tokenizer_path = f"{base_path}/{tgt_tokenizer_filename}"
        if os.path.isfile(src_tokenizer_path) and os.path.isfile(tgt_tokenizer_path):
            if src_tokenizer_filename not in repo_files and tgt_tokenizer_filename not in repo_files:
                print("Uploading tokenizers...")
                api.upload_file(
                    path_or_fileobj=src_tokenizer_path,
                    path_in_repo=src_tokenizer_filename,
                    repo_id=repo_id,
                    token=hf_token,
                )
                api.upload_file(
                    path_or_fileobj=tgt_tokenizer_path,
                    path_in_repo=tgt_tokenizer_filename,
                    repo_id=repo_id,
                    token=hf_token,
                )
        else:
            raise FileNotFoundError("One the tokenizers file is not present")

    # Load safetensor to provide a safer alternative to PyTorch pickle format
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_safetensor_path = os.path.join(tmp_dir, "model.safetensors")

        model = Transformer(
            src_vocab_size=config.src_tokenizer_vocab_size,
            tgt_vocab_size=config.tgt_tokenizer_vocab_size,
            num_encoder_blocks=config.model_configs[config.chosen_model_size].num_encoder_layers,
            num_decoder_blocks=config.model_configs[config.chosen_model_size].num_decoder_layers,
            d_model=config.model_configs[config.chosen_model_size].d_model,
            num_heads=config.model_configs[config.chosen_model_size].num_heads,
            d_ff=config.model_configs[config.chosen_model_size].d_ff,
            norm_type=config.model_configs[config.chosen_model_size].norm_type,
            dropout_prob=config.model_parameters.dropout,
            max_seq_len=config.tokenizer.max_seq_len,
        )

        model_state_dict = checkpoint["model"]
        first_layer_key = next(iter(model_state_dict.keys()))
        if first_layer_key.startswith("_orig_mod."):
            new_state_dict = {}
            prefix = "_orig_mod."
            for k, v in model_state_dict.items():
                new_k = k[len(prefix) :] if k.startswith(prefix) else k
                new_state_dict[new_k] = v
            model_state_dict = new_state_dict

        model.load_state_dict(model_state_dict, strict=True, assign=True)

        save_file(model.state_dict(), tmp_safetensor_path)

        if "model.safetensors" not in repo_files:
            api.upload_file(
                path_or_fileobj=tmp_safetensor_path,
                path_in_repo="model.safetensors",
                repo_id=repo_id,
                token=hf_token,
            )

        print("Successfully uploaded safetensor model.")

    # Clean config
    config.pop("wandb_organization", None)
    config.pop("output_dir", None)
    config.pop("base_path", None)
    config.pop("cache_ds_path", None)
    config.pop("s3_bucket_name", None)
    config.pop("ckpt_path_to_resume_from", None)
    config.pop("tokenizers_resume_path", None)

    # Upload config
    with tempfile.NamedTemporaryFile(suffix=".yaml") as tmp:
        tmp_config_path = tmp.name

        with open(tmp_config_path, "w+") as f:
            OmegaConf.save(config, f)

        print("Uploading config...")
        api.upload_file(
            path_or_fileobj=tmp_config_path, path_in_repo="config-lock.yaml", repo_id=repo_id, token=hf_token
        )

    # Generate and upload README
    print("Generating README...")
    readme_content = generate_readme(config, repo_id.split("/")[-1])

    with tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".md") as tmp:
        tmp.write(readme_content)
    try:
        tmp_readme_path = tmp.name

        print("Uploading README...")
        api.upload_file(path_or_fileobj=tmp_readme_path, path_in_repo="README.md", repo_id=repo_id, token=hf_token)
    finally:
        os.remove(tmp.name)

    print(f"\nSuccessfully uploaded all files to: https://huggingface.co/{repo_id}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Upload PyTorch checkpoint to Hugging Face Hub")
    parser.add_argument("checkpoint_path", type=str, help="Path to checkpoint file (local path or s3://bucket/key)")
    parser.add_argument(
        "--repo_id", type=str, default="giovo17/tfs-mt", help="Hugging Face repository ID (e.g., username/model-name)"
    )
    parser.add_argument("--private", default=True, help="Create a private repository")

    args = parser.parse_args()

    hf_token = os.getenv("HF_TOKEN")

    upload_to_huggingface(
        checkpoint_path=args.checkpoint_path,
        repo_id=args.repo_id,
        hf_token=hf_token,
        private=args.private,
    )



# Ignite custom checkpointer
#
# This file is derived from PyTorch Ignite (BSD 3-Clause License)
# Copyright (c) 2018-present, PyTorch-Ignite team
# License for original code: BSD 3-Clause
#
# Modifications:
# Author: Giovanni Spadaro - https://giovannispadaro.it
# Project: https://github.com/Giovo17/tfs-mt
# Documentation: https://giovo17.github.io/tfs-mt
#
# Copyright (c) Giovanni Spadaro.
# License for modifications: MIT
#
# The full text of both licenses can be found in the LICENSE file in the root directory of this source tree.

import os
import tempfile
from collections.abc import Callable, Mapping
from pathlib import Path
from typing import Any

import boto3
import botocore
import botocore.exceptions
import ignite.distributed as idist
import torch
from botocore.client import BaseClient
from dotenv import load_dotenv
from ignite.handlers.checkpoint import BaseSaveHandler


class BucketNotFoundError(Exception):
    def __init__(self, bucket):
        msg = f"S3 bucket '{bucket}' does not exist."
        super().__init__(msg)


class BucketNotEmptyError(Exception):
    def __init__(self, bucket, prefix, existing):
        msg = f"""Bucket '{bucket}' already contains .pt or .pth objects under prefix "
        '{prefix}': {existing}. Set `require_empty=False` to ignore."""
        super().__init__(msg)


class S3Saver(BaseSaveHandler):
    """
    Handler that saves a checkpoint directly to an S3 bucket.

    Args:
        bucket:      Name of the target S3 bucket.
        prefix:      Optional key prefix (e.g. ``"checkpoints/"``).  The final
                     object key will be ``prefix + filename``.
        atomic:      If ``True`` the checkpoint is first written to a temporary
                     local file and then uploaded, guaranteeing that a partially
                     uploaded object never appears in the bucket.
        create_bucket: If ``True`` the bucket will be created when it does not
                     already exist (requires appropriate IAM permissions).
        require_empty: If ``True`` raises an error when the bucket already
                     contains objects with the same ``prefix`` and ``.pt`` suffix.
        save_on_rank: Rank on which the checkpoint will be saved (useful for
                     distributed training).
        aws_kwargs:  Additional keyword arguments passed to ``boto3.client``.
        **kwargs:    Keyword arguments forwarded to ``torch.save`` (or ``xm.save``
                     for XLA devices).
    """

    def __init__(
        self,
        bucket: str,
        prefix: str = "",
        atomic: bool = True,
        require_empty: bool = True,
        save_on_rank: int = 0,
        **kwargs: Any,
    ):
        self.bucket = bucket
        self.prefix = prefix.rstrip("/") + "/" if prefix and not prefix.endswith("/") else prefix
        self._atomic = atomic
        self.save_on_rank = save_on_rank
        self.kwargs = kwargs

        # Check if any required env vars are missing, load .env if needed
        required_vars = ["AWS_ACCESS_KEY", "AWS_SECRET_KEY", "AWS_REGION", "CUSTOM_BASE_URL"]
        if any(os.getenv(var) is None for var in required_vars):
            load_dotenv()
        self.s3 = self._make_s3_client()

        if idist.get_rank() == save_on_rank:
            self._check_and_setup(require_empty)

    @staticmethod
    def _make_s3_client() -> BaseClient:
        """Create a boto3 S3 client using env vars and optional custom endpoint."""
        access_key = os.getenv("AWS_ACCESS_KEY")
        secret_key = os.getenv("AWS_SECRET_KEY")
        region = os.getenv("AWS_REGION")
        custom_base = os.getenv("CUSTOM_BASE_URL")

        session_kwargs = {}
        if access_key and secret_key:
            session_kwargs["aws_access_key_id"] = access_key
            session_kwargs["aws_secret_access_key"] = secret_key
        if region:
            session_kwargs["region_name"] = region

        session = boto3.Session(**session_kwargs)

        client_kwargs: dict = {}
        if custom_base:
            # e.g. CUSTOM_BASE_URL = "example.com" → endpoint https://s3.example.com
            client_kwargs["endpoint_url"] = f"https://s3.{custom_base}"

        return session.client("s3", **client_kwargs)

    def _check_and_setup(self, require_empty: bool) -> None:
        """Validate bucket existence and optionally create it."""

        # Verify bucket existence
        try:
            self.s3.head_bucket(Bucket=self.bucket)
        except botocore.exceptions.ClientError as exc:
            if exc.response["Error"]["Code"] == "404":
                raise BucketNotFoundError(self.bucket) from exc
            else:
                raise

        if require_empty:
            resp = self.s3.list_objects_v2(Bucket=self.bucket, Prefix=self.prefix)
            existing = [obj["Key"] for obj in resp.get("Contents", []) if obj["Key"].endswith((".pt", ".pth"))]
            if existing:
                raise BucketNotEmptyError(self.bucket, self.prefix, existing)

    def __call__(
        self,
        checkpoint: Mapping,
        filename: str,
        metadata: Mapping | None = None,
    ) -> None:
        """Save ``checkpoint`` to S3 under ``self.prefix + filename``."""
        s3_key = f"{self.prefix}{filename}"

        if idist.has_xla_support:
            import torch_xla.core.xla_model as xm

            self._save_func(checkpoint, s3_key, xm.save)
        elif self.save_on_rank == idist.get_rank():
            self._save_func(checkpoint, s3_key, torch.save)

    def _save_func(self, checkpoint: Mapping, s3_key: str, func: Callable) -> None:
        """Write checkpoint to a temporary file (if atomic) and upload to S3."""
        if not self._atomic:
            # Direct upload from an inmemory buffer (no temp file)
            with tempfile.SpooledTemporaryFile() as buf:
                func(checkpoint, buf, **self.kwargs)
                buf.seek(0)
                self.s3.upload_fileobj(buf, self.bucket, s3_key)
        else:
            # Write to a local temp file first, guarantees that a partially written object never appears in the bucket.
            with tempfile.NamedTemporaryFile(delete=False, dir="/tmp") as tmp:
                tmp_path = Path(tmp.name)
                try:
                    func(checkpoint, tmp_path, **self.kwargs)
                except BaseException:
                    tmp.close()
                    os.remove(tmp_path)
                    raise

            # Upload the completed file
            self.s3.upload_file(str(tmp_path), self.bucket, s3_key)

            # Clean up the temporary file
            os.remove(tmp_path)

            # Optional: make the object readable by others (mirrors chmod)
            # Users can control this via an env var if desired.
            if os.getenv("S3_PUBLIC_READ") == "1":
                self.s3.put_object_acl(Bucket=self.bucket, Key=s3_key, ACL="public-read")

    def remove(self, filename: str) -> None:
        """Delete ``filename`` from the bucket."""
        if idist.get_rank() == self.save_on_rank:
            s3_key = f"{self.prefix}{filename}"
            self.s3.delete_object(Bucket=self.bucket, Key=s3_key)



# Transformer decoding utils
#
# Author: Giovanni Spadaro - https://giovannispadaro.it
# Project: https://github.com/Giovo17/tfs-mt
# Documentation: https://giovo17.github.io/tfs-mt
#
# Copyright (c) Giovanni Spadaro.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in the root directory of this source tree.

import typing

import numpy as np
import torch

from .architecture import Transformer
from .data_utils import BaseTokenizer


@torch.inference_mode()
def greedy_decoding(
    model: Transformer,
    tgt_tokenizer: BaseTokenizer,
    src_tokens: np.ndarray | torch.Tensor,
    src_mask: np.ndarray | torch.Tensor,
    max_target_tokens: int = 128,
    output_mode: str = "str",
) -> list[str] | list[list[str]]:
    """
    Supports batch (decode multiple source sentences) greedy decoding.

    Note: Example
        We input `<s>` and do a forward pass. We get intermediate activations for `<s>` and at the output at position
        0, after the doing linear layer we get e.g. token `<I>`. Now we input `<s>`,`<I>` but `<s>`'s activations will remain
        the same. Similarly say we now got `<am>` at output position 1, in the next step we input `<s>`,`<I>`,`<am>` and so `<I>`'s
        activations will remain the same as it only looks at/attends to itself and to `<s>` and so forth.

    Note:
        Decoding could be further optimized to cache old token activations because they can't look ahead and so
        adding a newly predicted token won't change old token's activations.

    Args:
        model (Transformer): Encoder-decoder translation model.
        tgt_tokenizer (BaseTokenizer): Target text tokenizer.
        src_tokens (torch.Tensor): Source tokens.
        src_mask (torch.Tensor): Source tokens mask.
        max_target_tokens (int, optional): Max target tokens to output. Defaults to 128.
        output_mode (str, optional): Output mode, if `str` it will return the sequence as a string, if `tokens` it will return as list of tokens. Defaults to "str".

    Returns:
        list[str] | list[list[str]]: Decoded sequences as strings or as list of tokens.
    """

    device = next(model.parameters()).device

    sos_token = tgt_tokenizer.decode([tgt_tokenizer.sos_token_idx])[0]
    eos_token = tgt_tokenizer.decode([tgt_tokenizer.eos_token_idx])[0]

    # Get encoder representation
    if isinstance(src_tokens, np.ndarray) and isinstance(src_tokens, np.ndarray):
        src_tokens = torch.tensor(src_tokens)
        src_mask = torch.tensor(src_mask)
    if src_tokens.ndim == 1 and src_mask.ndim == 1:
        src_tokens.unsqueeze_(0)
        src_mask.unsqueeze_(0)

    src_tokens = src_tokens.to(device)
    src_mask = src_mask.to(device)

    encoder_representation = model.encode(src_tokens, src_mask)

    # Generate a batch of sequences starting with SOS token, batch size is inferred by the encoder representation tensor
    tgt_sequence_batch_text = [[sos_token] for _ in range(encoder_representation.shape[0])]
    tgt_sequence_batch = torch.tensor(
        [[tgt_tokenizer.sos_token_idx] for _ in range(encoder_representation.shape[0])], device=device
    )

    # This list handles when to stop the tokens generation for each sequence in the batch
    is_decoded = [False] * encoder_representation.shape[0]

    while True:
        tgt_mask = typing.cast(torch.Tensor, tgt_tokenizer.encode(tgt_sequence_batch, return_only_mask=True))

        # Due to cross attention max tgt sequences cannot be longer than max src sequences
        if tgt_sequence_batch.shape[1] > encoder_representation.shape[1]:
            dummy_tensor = torch.ones_like(encoder_representation, device=encoder_representation.device)
            dummy_tensor = dummy_tensor[:, 0, :].unsqueeze(1)
            encoder_representation = torch.cat((encoder_representation, dummy_tensor), dim=1)

            addon_mask = torch.zeros((src_mask.shape[0], 1), dtype=torch.bool, device=src_mask.device)
            src_mask = torch.cat((src_mask, addon_mask), dim=1)

        # Shape = (B*T, V) where T is the current token-sequence length and V target vocab size
        decoder_output = model.decode(tgt_sequence_batch, encoder_representation, tgt_mask, src_mask)

        # Extract only the indices of last token for every target sentence
        num_of_tgt_tokens = tgt_sequence_batch.shape[1]
        decoder_output = decoder_output[:, num_of_tgt_tokens - 1 :: num_of_tgt_tokens]

        # Greedy decode tokens selecting the most probable one and discard other tokens
        most_probable_last_token_indices = torch.argmax(decoder_output, dim=-1).cpu().numpy()

        # Find target tokens associated with these indices
        predicted_words = []
        for row in most_probable_last_token_indices:
            predicted_words.append(tgt_tokenizer.decode(row)[0])

        for idx, predicted_word in enumerate(predicted_words):
            tgt_sequence_batch_text[idx].append(predicted_word)
            # Once EOS token is generated for a sentence in the batch it gets flagged in is_decoded list
            if predicted_word == eos_token:
                is_decoded[idx] = True

        if all(is_decoded) or num_of_tgt_tokens == max_target_tokens:
            break

        # Prepare the input for the next iteration: merge old token ids with the new column of most probable token ids
        tgt_sequence_batch = torch.cat(
            (tgt_sequence_batch, torch.tensor(most_probable_last_token_indices, device=device)), dim=1
        )

    # Post process the sentences: remove everything after the EOS token
    post_processed_sequences = []
    for tgt_sequence in tgt_sequence_batch_text:
        try:
            target_index = tgt_sequence.index(eos_token) + 1
        except ValueError:
            target_index = None

        tgt_sequence = tgt_sequence[:target_index]
        post_processed_sequences.append(tgt_sequence)

    if output_mode == "str":
        post_processed_sequences_str = []
        for i in range(len(post_processed_sequences)):
            seq = post_processed_sequences[i]
            seq.remove(sos_token)
            if eos_token in seq:
                seq.remove(eos_token)
            sequence_as_str = " ".join(seq)
            post_processed_sequences_str.append(sequence_as_str)
        post_processed_sequences = post_processed_sequences_str

    return post_processed_sequences


@torch.inference_mode()
def beam_decoding(
    model: Transformer,
    tgt_tokenizer: BaseTokenizer,
    src_tokens: np.ndarray | torch.Tensor,
    src_mask: np.ndarray | torch.Tensor,
    max_target_tokens: int = 128,
    output_mode: str = "str",
) -> list[str]:
    """
    TBA
    """

    raise NotImplementedError()



# Transformer training utils
#
# Author: Giovanni Spadaro - https://giovannispadaro.it
# Project: https://github.com/Giovo17/tfs-mt
# Documentation: https://giovo17.github.io/tfs-mt
#
# Copyright (c) Giovanni Spadaro.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in the root directory of this source tree.

import math
import os
import tempfile
from collections.abc import Callable, Iterable, Mapping
from contextlib import suppress
from datetime import datetime
from logging import Logger
from typing import Any, cast

import botocore
import botocore.exceptions
import torch
import torch.nn.functional as F
from ignite.engine import DeterministicEngine, Engine, Events
from ignite.handlers import Checkpoint, DiskSaver, ProgressBar, global_step_from_engine
from ignite.handlers.checkpoint import BaseSaveHandler
from ignite.handlers.early_stopping import EarlyStopping
from ignite.handlers.time_limit import TimeLimit
from ignite.metrics import GpuInfo
from ignite.metrics.metric import Metric
from omegaconf import OmegaConf
from omegaconf.dictconfig import DictConfig
from omegaconf.listconfig import ListConfig
from torch import nn
from torch.amp import GradScaler, autocast
from torch.optim.lr_scheduler import LRScheduler
from torch.optim.optimizer import Optimizer

from .data_utils import BaseTokenizer, WordTokenizer
from .ignite_custom_utils.checkpoint import BucketNotFoundError, S3Saver
from .ignite_custom_utils.common import setup_trackio_logging
from .ignite_custom_utils.trackio_logger import TrackioLogger
from .ignite_custom_utils.wandb_logger import WandBLogger


class CheckpointNotFoundError(Exception):
    def __init__(self, checkpoint_path):
        msg = f"Given {checkpoint_path!s} does not exist."
        super().__init__(msg)


class InvalidCheckpointS3PathError(Exception):
    def __init__(self, msg="checkpoint_s3_path must be of form s3://bucket/key"):
        super().__init__(msg)


class S3FailedDownloadError(Exception):
    def __init__(self, checkpoint_path, e):
        msg = f"Failed to download checkpoint from S3 ({checkpoint_path!s}): {e!s}"
        super().__init__(msg)


class DimensionError(Exception):
    def __init__(self, vocab_size, x_shape_1):
        msg = f"Dimension error, the second dimension should be {vocab_size}, got {x_shape_1}"
        super().__init__(msg)


def resume_from_ckpt(
    checkpoint_path: str,
    to_load: Mapping | None = None,
    device: torch.device | None = None,
    logger: Logger | None = None,
    strict: bool = True,
    resume_tokenizers: bool = False,
    tokenizers_type: str | None = "word",
) -> None | tuple[BaseTokenizer, BaseTokenizer]:
    """Loads state dict from a checkpoint file to resume the training or loads tokenizers.
    It supports loading from local or bucket s3 checkpoint.

    Args:
        checkpoint_path (str): Path to the checkpoint file (or folder) in S3 bucket or in filesystem.
        to_load (Mapping | None, optional): A dictionary with objects.. Defaults to None.
        device (torch.device | None, optional): Device. Defaults to None.
        logger (Logger | None, optional): To log info about resuming from a checkpoint. Defaults to None.
        strict (bool, optional): Whether to strictly enforce that the keys in `state_dict` match the keys returned by this module's `state_dict()` function. Defaults to True.
        resume_tokenizers (bool, optional): Whether to load only tokenizers. Defaults to False.
        tokenizers_type (str, optional): Tokenizers type (Word, BPE).

    Raises:
        CheckpointNotFoundError: Raised when checkpoint file doesn't exist.
        InvalidCheckpointS3PathError: Raised when bucket and file key are not correctly extracted from provided url.
        S3FailedDownloadError: Raised when download fails.

    Returns:
        None | tuple[BaseTokenizer, BaseTokenizer]: Pretrained tokenizers if resume_tokenizers. Otherwise None.
    """

    resume_method = "local"
    if checkpoint_path.startswith("s3://"):
        resume_method = "bucket-s3"

    if resume_method == "local":
        if not os.path.isfile(checkpoint_path):
            raise CheckpointNotFoundError(checkpoint_path)

        if resume_tokenizers:
            # Detect if the function checkpoint_path is the folder/bucket path or the pt export filepath
            ckpt_basepath = (
                "/".join(checkpoint_path.split("/")[:-1])
                if checkpoint_path.endswith((".pt", ".pth"))
                else checkpoint_path
            )
            src_tokenizer = WordTokenizer.from_pretrained(ckpt_basepath + f"/src_tokenizer_{tokenizers_type}.json")
            tgt_tokenizer = WordTokenizer.from_pretrained(ckpt_basepath + f"/tgt_tokenizer_{tokenizers_type}.json")
            return src_tokenizer, tgt_tokenizer

        checkpoint = torch.load(checkpoint_path, map_location=device, weights_only=True)

        if to_load is not None:
            Checkpoint.load_objects(to_load=to_load, checkpoint=checkpoint, strict=strict)
        if logger is not None:
            logger.info("Successfully resumed from a local checkpoint: %s", checkpoint_path)

    else:  # resume_method == "bucket-s3":
        _, _, path = checkpoint_path.partition("s3://")
        bucket, _, key = path.partition("/")

        if bucket == "" or key == "":
            raise InvalidCheckpointS3PathError()

        src_tokenizer_key = key.split("/")[0] + f"/src_tokenizer_{tokenizers_type}.json" if resume_tokenizers else None
        tgt_tokenizer_key = key.split("/")[0] + f"/tgt_tokenizer_{tokenizers_type}.json" if resume_tokenizers else None

        s3 = S3Saver._make_s3_client()

        if resume_tokenizers:  # Download tokenizers to temperary files
            with (
                tempfile.NamedTemporaryFile(delete=False) as src_tmp,
                tempfile.NamedTemporaryFile(delete=False) as tgt_tmp,
            ):
                src_tmp_path = src_tmp.name
                tgt_tmp_path = tgt_tmp.name
            try:
                try:
                    s3.download_file(bucket, src_tokenizer_key, src_tmp_path)
                    s3.download_file(bucket, tgt_tokenizer_key, tgt_tmp_path)
                except Exception as e:
                    raise S3FailedDownloadError(checkpoint_path, e) from e

                src_tokenizer = WordTokenizer.from_pretrained(src_tmp_path)
                tgt_tokenizer = WordTokenizer.from_pretrained(tgt_tmp_path)
                logger.info(f"Successfully resumed tokenizers from a bucket s3 checkpoint: {bucket}")
                return src_tokenizer, tgt_tokenizer
            finally:
                with suppress(OSError):
                    os.remove(src_tmp_path)
                    os.remove(tgt_tmp_path)

        else:  # Download to_load to a temporary file
            with tempfile.NamedTemporaryFile(delete=False) as tmp:
                tmp_path = tmp.name
            try:
                try:
                    s3.download_file(bucket, key, tmp_path)
                except Exception as e:
                    raise S3FailedDownloadError(checkpoint_path, e) from e

                checkpoint = torch.load(tmp_path, map_location=device, weights_only=True)

                # Fix model weights loading when the checkpoint layers' keys start with "_orig_mod."
                model_state_dict = checkpoint["model"]
                first_layer_key = next(iter(model_state_dict.keys()))
                if first_layer_key.startswith("_orig_mod."):
                    new_state_dict = {}
                    prefix = "_orig_mod."
                    for k, v in model_state_dict.items():
                        new_k = k[len(prefix) :] if k.startswith(prefix) else k
                        new_state_dict[new_k] = v
                    checkpoint["model"] = new_state_dict

                if to_load is not None:
                    Checkpoint.load_objects(to_load=to_load, checkpoint=checkpoint, strict=strict)

                if logger is not None:
                    logger.info(f"Successfully resumed training objects from a bucket s3 checkpoint: {checkpoint_path}")
            finally:
                with suppress(OSError):
                    os.remove(tmp_path)


def setup_output_dir(config: DictConfig | ListConfig) -> str:
    """Create output folder."""
    output_dir = config.output_dir

    now = datetime.now().strftime("%Y%m%d-%H%M%S")
    name = f"{now}-min_lr-{config.training_hp.lr_scheduler.min_lr}-max_lr-{config.training_hp.lr_scheduler.max_lr}"
    output_dir = os.path.join(config.output_dir, name)
    os.makedirs(output_dir, exist_ok=True)

    return output_dir


def s3_upload(filepath: str, bucket: str, s3_key: str | None = None) -> None:
    """Upload a file on S3 bucket."""

    if not os.path.exists(filepath):
        print("File was not found at the following path, skipping the upload")
        print(filepath)
        return None

    s3 = S3Saver._make_s3_client()
    s3_key = s3_key or filepath.split("/")[-1]  # Default key = filename

    # Verify bucket existence
    try:
        s3.head_bucket(Bucket=bucket)
    except botocore.exceptions.ClientError as exc:
        print(f"Bucket not found - {bucket}")
        if exc.response["Error"]["Code"] == "404":
            raise BucketNotFoundError(bucket) from exc
        else:
            raise

    try:
        s3.upload_file(filepath, bucket, s3_key)
        print(f"Uploaded '{filepath}' to s3://{bucket}/{s3_key}")
    except Exception as exc:
        print(f"Failed to upload '{filepath}' to S3: {exc}")
        raise


def save_config(
    config: DictConfig | ListConfig,
    output_dir: str,
    enable_ckpt: bool = True,
):
    """Save configuration to config-lock.yaml for result reproducibility."""
    with open(f"{output_dir}/config-lock.yaml", "w+") as f:
        OmegaConf.save(config, f)
    # Upload to S3 endpoint
    if config.s3_bucket_name is not None and enable_ckpt:
        s3_upload(
            filepath=f"{output_dir}/config-lock.yaml",
            bucket=config.s3_bucket_name,
            s3_key=f"{config.model_name}/config-lock.yaml",
        )


def log_metrics(engine: Engine, tag: str) -> None:
    """Log `engine.state.metrics` with given `engine` and `tag`.

    Args:
        engine (Engine): Instance of `Engine` which metrics to log.
        tag (str): A string to add at the start of output.
    """

    metrics_format = f"{tag} [{engine.state.epoch}/{engine.state.iteration}]: {engine.state.metrics}"
    engine.logger.info(metrics_format)


def setup_exp_logging(
    config: DictConfig | ListConfig,
    trainer: Engine,
    optimizer: Optimizer | dict[str, Optimizer],
    evaluator: Engine | dict[str, Engine],
    metrics: dict[str, Metric],
    model: nn.Module | None = None,
    return_all_loggers: bool = False,
) -> WandBLogger | tuple[WandBLogger, TrackioLogger]:
    """Setup Experiment Tracking with WandB and Trackio loggers.

    Using `common.setup_wandb_logging` which setup an ignite's Engine compatible WandB logger.
    It takes as kwargs `wandb.init` compatible arguments.
    Same for `trackio`.

    Note: References
        1. `setup_wandb_logging` documentation page [[link](https://docs.pytorch.org/ignite/contrib/engines.html#ignite.contrib.engines.common.setup_wandb_logging)]
        2. `WandBLogger` documentation page [[link](https://docs.pytorch.org/ignite/generated/ignite.handlers.wandb_logger.html#ignite.handlers.wandb_logger.WandBLogger)]
        3. `wandb.init` documentation page [[link](https://docs.wandb.ai/models/ref/python/functions/init)]
        4. `trackio.init` documentation page [[link](https://huggingface.co/docs/trackio/en/api#trackio.init)]
    """

    wandb_logger = WandBLogger(
        entity=config.wandb_organization,
        project=config.model_base_name,
        name=config.model_name,
        config=config._content,
        tags=["pytorch", "nlp", "machine-translation"],
    )

    wandb_logger.attach_output_handler(
        trainer,
        event_name=Events.ITERATION_COMPLETED(every=config.log_every_iters),
        tag="training",
        # metric_names=list(trainer.state.metrics.keys()),  # Data recorded during trainig: loss, number of seen tokens
        output_transform=lambda loss: loss,
        # output_transform=lambda loss: {"Loss": loss}
    )

    wandb_logger.attach_output_handler(
        cast(Engine, evaluator),
        event_name=Events.EPOCH_COMPLETED(every=1),
        tag="test_eval",
        metric_names=list(metrics.keys()),
        global_step_transform=lambda *_: trainer.state.iteration,
    )

    wandb_logger.attach_opt_params_handler(
        trainer, event_name=Events.ITERATION_STARTED(every=config.log_every_iters), optimizer=optimizer, param_name="lr"
    )

    if model is not None:  # Watch model to log gradients
        wandb_logger.watch(model)

    if not return_all_loggers:
        return wandb_logger

    # Custom trackio directory https://huggingface.co/docs/trackio/en/environment_variables#trackiodir
    os.environ["TRACKIO_DIR"] = config.output_dir + "/trackio"

    trackio_logger = setup_trackio_logging(
        trainer,
        optimizer,
        evaluator,
        config.log_every_iters,
        # trackio.init kwargs
        project=config.model_base_name,
        name=config.model_name,
        config=config._content,
    )
    return wandb_logger, trackio_logger


def get_device():
    if torch.cuda.is_available():
        return torch.device("cuda:0")
    elif torch.mps.is_available():
        return torch.device("mps:0")
    else:
        return torch.device("cpu")


def setup_handlers(
    trainer: Engine,
    evaluator: Engine,
    config: DictConfig | ListConfig,
    to_save_train: dict | None = None,
    to_save_test: dict | None = None,
    enable_ckpt: bool = True,
) -> None:
    """Setup Ignite handlers."""

    if enable_ckpt:  # Setup checkpointing
        # Setup checkpoints savers
        disk_saver = DiskSaver(os.path.join(config.output_dir, "checkpoints"), require_empty=False)
        s3_saver = (
            S3Saver(bucket=config.s3_bucket_name, prefix=config.model_name + "/")
            if config.s3_bucket_name is not None
            else None
        )

        # Training checkpointing.
        # Do it locally only if s3 checkpointing is disabled to save disk space in cloud instance.
        to_save_train = to_save_train or {}
        ckpt_handler_train = Checkpoint(
            to_save_train,
            save_handler=cast(BaseSaveHandler, s3_saver if config.s3_bucket_name is not None else disk_saver),
            filename_prefix=config.model_base_name,
            n_saved=config.checkpoints_retain_n,
        )
        train_ckpt_events = (
            Events.ITERATION_COMPLETED(every=config.save_every_iters)
            | Events.EPOCH_COMPLETED(every=1)
            | Events.TERMINATE
            | Events.COMPLETED
        )
        trainer.add_event_handler(train_ckpt_events, ckpt_handler_train)

        # Evaluation checkpointing.
        # Do it locally only if s3 checkpointing is disabled to save disk space in cloud instance.
        global_step_transform = None
        if to_save_train.get("trainer", None) is not None:
            global_step_transform = global_step_from_engine(to_save_train["trainer"])
        ckpt_handler_test = Checkpoint(
            to_save_test,
            save_handler=s3_saver if config.s3_bucket_name is not None else disk_saver,
            filename_prefix="best",
            n_saved=config.checkpoints_retain_n,
            global_step_transform=global_step_transform,
            score_name="test_bleu",
            score_function=Checkpoint.get_default_score_fn("Bleu"),
        )
        test_ckpt_events = Events.EPOCH_COMPLETED(every=1) | Events.TERMINATE | Events.COMPLETED
        evaluator.add_event_handler(test_ckpt_events, ckpt_handler_test)

    # Time limit reached policy to stop training. Mainly used in Kaggle due to 12 hours run limit.
    if config.time_limit_sec != -1:
        print(f"Setting up training time limit to {int(config.time_limit_sec) / 3600} hours.")
        trainer.add_event_handler(Events.ITERATION_COMPLETED, TimeLimit(config.time_limit_sec))

    # Iterations and epochs progress bars
    ProgressBar(persist=True, bar_format="").attach(
        trainer, event_name=Events.EPOCH_STARTED, closing_event_name=Events.COMPLETED
    )
    ProgressBar(persist=False).attach(
        trainer, metric_names="all", event_name=Events.ITERATION_COMPLETED(every=config.update_pbar_every_iters)
    )

    if torch.cuda.is_available():
        GpuInfo().attach(trainer, name="gpu")


def setup_early_stopping(
    trainer: Engine,
    evaluator: Engine,
    config: DictConfig | ListConfig,
) -> None:
    """Setup early stopping."""

    def score_fn(engine: Engine):
        return engine.state.metrics["Bleu"]

    es = EarlyStopping(
        patience=config.training_hp.early_stopping.patience,  # Considered in number of iterations
        score_function=score_fn,
        trainer=trainer,
        min_delta=config.training_hp.early_stopping.min_delta,
    )
    evaluator.add_event_handler(Events.ITERATION_COMPLETED, es)


def nlp_metric_transform(
    output: tuple[torch.Tensor, torch.Tensor], tgt_tokenizer: BaseTokenizer
) -> tuple[list[list[str]], list[list[list[str]]]]:
    """Transform `eval_one_iter` output to be compliant with ignite nlp metrics.

    Note: References
        1. Bleu documentation page [[link](https://docs.pytorch.org/ignite/generated/ignite.metrics.Bleu.html)]
        2. Rouge documentation page [[link](https://docs.pytorch.org/ignite/generated/ignite.metrics.Rouge.html)]

    Args:
        output (tuple[torch.Tensor, torch.Tensor]): Output of `eval_one_iter`.
        tgt_tokenizer (BaseTokenizer): Target tokenizer used to decode tokens.

    Returns:
        tuple[list[list[str]], list[list[list[str]]]]: Metrics complatible output.
    """

    output_logits, tgt_output_label = output

    # Get predicted tokens from logits
    output_logits = output_logits.detach()
    output_tokens = torch.argmax(output_logits, dim=-1)

    # Move to list of int for tokenizer.decode compatibility
    output_tokens = output_tokens.cpu().numpy().tolist()
    tgt_output_label = tgt_output_label.cpu().numpy().tolist()

    # Decode batched token sequences to lists of lists of vocab token sequences
    y_pred = [tgt_tokenizer.decode(sample) for sample in output_tokens]
    y = [tgt_tokenizer.decode(sample) for sample in tgt_output_label]

    # Adjust shape. Ignite wants a corpus of lists of target label sentences for each hypotheses.
    # Since the dataset proposes only one target translation for a given input, y is wrapped in a list.
    y = [y]

    return y_pred, y


def loss_metric_transform(
    output: tuple[torch.Tensor, torch.Tensor],
    loss_type: str,
) -> tuple[torch.Tensor, torch.Tensor]:
    """Transform `eval_one_iter` output to be compliant with torch loss computation.

    Args:
        output (tuple[torch.Tensor, torch.Tensor]): Output of `eval_one_iter`.
        loss_type (str): Loss type choosen.

    Returns:
        tuple[torch.Tensor, torch.Tensor]: Loss compatible output.
    """

    output_logits, tgt_output_label = output

    if loss_type == "crossentropy":
        pred = output_logits
    elif loss_type == "KLdiv-labelsmoothing":
        pred = F.log_softmax(output_logits, dim=-1)
    else:
        raise ValueError(f"Invalid loss type, got {loss_type}")

    return (pred.reshape(-1, pred.size(-1)), tgt_output_label.reshape(-1))


def get_param_groups(
    model_named_parameters: Iterable[tuple[str, nn.Parameter]], weight_decay: float
) -> list[dict[str, Any]]:
    """Create parameter groups for optimizer with selective weight decay application.

    Parameters excluded from weight decay:

    - Bias terms
    - Layer normalization weights
    - Embedding weights

    Args:
        model_named_parameters (Iterable[tuple[str, nn.Parameter]]): Model's named parameters.
        weight_decay (float): Weight decay coefficient to apply to parameters as mentioned.

    Returns:
        list[dict[str, Any]]: List of two parameter group dictionaries to be passed to the optimizer.
    """

    decay = []
    no_decay = []
    for name, param in model_named_parameters:
        if not param.requires_grad:
            continue
        # Exclude biases, LayerNorm and embeddings from weight decay regularization
        if name.endswith(".bias") or "layer_norm" in name or "embeddings" in name:
            no_decay.append(param)
        else:
            decay.append(param)

    return [
        {"params": decay, "weight_decay": weight_decay},
        {"params": no_decay, "weight_decay": 0.0},
    ]


def setup_lr_lambda_fn(config: DictConfig | ListConfig) -> Callable[[int], float]:
    """Setup function that will govern the learning rate scheduling.

    It supports two scheduling strategies: Warmup-Stable-Decay (WSD) and the original lr scheduling from Attention Is All You Need.

    Args:
        config (DictConfig | ListConfig): Project config file.

    Returns:
        Callable[[int], float]: Lambda function mapping step number to LR multiplier.

    Raises:
        ValueError: Raised when config.training_hp.lr_scheduler.type is not "wsd" nor "original".
    """

    min_lr = config.training_hp.lr_scheduler.min_lr
    max_lr = config.training_hp.lr_scheduler.max_lr

    total_iters = config.num_train_iters_per_epoch * config.training_hp.num_epochs
    warmup_iters = config.training_hp.lr_scheduler.warmup_iters
    stable_iters = config.training_hp.lr_scheduler.stable_iters_prop * (total_iters - warmup_iters)
    decay_iters = total_iters - warmup_iters - stable_iters

    d_model = config.model_configs[config.chosen_model_size].d_model

    def wsd_lr_lambda(step):
        if step < warmup_iters:  # Warmup phase
            return max_lr * ((step + 1) / warmup_iters)

        elif step < stable_iters + warmup_iters:  # Stable phase
            return max_lr

        else:  # Decay phase
            step = step - warmup_iters - stable_iters
            cosine_decay = 0.5 * (1 + math.cos(math.pi * step / decay_iters))
            return min_lr + (max_lr - min_lr) * cosine_decay

    def original_lr_lambda(step):
        step = max(step, 1)  # Avoids zero division in step**-0.5
        return (d_model**-0.5) * min(step**-0.5, step * (warmup_iters**-1.5))

    if config.training_hp.lr_scheduler.type == "wsd":
        return wsd_lr_lambda
    elif config.training_hp.lr_scheduler.type == "original":
        return original_lr_lambda
    else:
        raise ValueError(f"Invalid learning rate scheduler type, got {config.training_hp.lr_scheduler.type}")


class KLDivLabelSmoothingLoss(nn.Module):
    def __init__(self, vocab_size: int, padding_idx: int, smoothing: float = 0.1):
        super().__init__()
        self.criterion = nn.KLDivLoss(reduction="sum")
        self.padding_idx = padding_idx
        self.confidence = 1.0 - smoothing
        self.smoothing = smoothing
        self.vocab_size = vocab_size

    def forward(self, x, target, return_true_dist: bool = False):
        if x.shape[1] != self.vocab_size:
            raise DimensionError(self.vocab_size, x.shape[1])

        true_dist = x.data.clone()
        true_dist.fill_(self.smoothing / (self.vocab_size - 2))
        true_dist.scatter_(1, target.data.unsqueeze(1), self.confidence)
        true_dist[:, self.padding_idx] = 0
        mask = torch.nonzero(target.data == self.padding_idx)
        if mask.dim() > 0:
            true_dist.index_fill_(0, mask.squeeze(), 0.0)

        if return_true_dist:
            return self.criterion(x, true_dist.clone().detach()), true_dist
        return self.criterion(x, true_dist.clone().detach())


def compute_grad_norm(params: Iterable[nn.Parameter]) -> float:
    total_norm = 0.0
    for p in params:
        param_norm = p.grad.data.norm(2)
        total_norm += math.pow(param_norm.item(), 2)
    return round(math.sqrt(total_norm), 3)


def setup_trainer(
    config: DictConfig | ListConfig,
    model: nn.Module,
    optimizer: Optimizer,
    lr_scheduler: LRScheduler,
    loss_fn: nn.Module,
    device: torch.device,
) -> Engine | DeterministicEngine:
    """Setup a trainer with mixed precision training support.

    Args:
        config (DictConfig | ListConfig): Project config file.
        model (nn.Module): Transformer model.
        optimizer (Optimizer): Optimizer.
        loss_fn (nn.Module): Loss function.
        device (torch.device): Device.

    Returns:
        Engine | DeterministicEngine: Trainer object.
    """

    # Gradient scaler for mixed precision training. Not required for bfloat16 training, cause it has the same range of float32.
    # It helps prevent gradients with small magnitudes from underflowing when training with mixed precision.
    # Reference: https://docs.pytorch.org/docs/stable/amp.html#gradient-scaling
    scaler = GradScaler(
        device.type,
        enabled=config.training_hp.use_amp and config.training_hp.amp_dtype not in ["float32", "bfloat16"],
    )

    amp_dtype_dict = {"float32": torch.float32, "bfloat16": torch.bfloat16, "float16": torch.float16}

    def train_one_iter(engine: Engine | DeterministicEngine, batch: dict[str, torch.Tensor | str]) -> dict[str, Any]:
        # non_blocking asynchronously transfers tensor from CPU to device. More here: https://docs.pytorch.org/tutorials/intermediate/pinmem_nonblock.html
        src_sequence = batch["src"].to(device, non_blocking=True, dtype=torch.long)
        tgt_sequence = batch["tgt"].to(device, non_blocking=True, dtype=torch.long)
        src_mask = batch["src_mask"].to(device, non_blocking=True, dtype=torch.long)
        # Mask is not shrinked accordingly to tgt_sequence here. It will be handled during attention processing.
        tgt_mask = batch["tgt_mask"].to(device, non_blocking=True, dtype=torch.long)  # [:, :-1]

        # Shifted target sequence as label for teacher forcing. Reshape to 1D tensor to later compute loss
        tgt_output_label = tgt_sequence[:, 1:]

        tgt_input_sequence = tgt_sequence[:, :-1]

        # Count how many tokens encoder and decoder see during training, excluding PAD, SOS and EOS tokens
        # NOTE The mask is implemented to include a token in the attention computation if the value at the corresponding index in the mask is True
        num_src_tokens = src_mask.to(torch.int8).sum().item() - 2 * src_mask.size(0)
        num_tgt_tokens = tgt_mask.to(torch.int8).sum().item() - 2 * tgt_mask.size(0)
        engine.state.tokens_seen_src = getattr(engine.state, "tokens_seen_src", 0) + num_src_tokens
        engine.state.tokens_seen_tgt = getattr(engine.state, "tokens_seen_tgt", 0) + num_tgt_tokens

        model.train()

        optimizer.zero_grad()

        # Mixed precision training if enabled in config. Reference: https://docs.pytorch.org/tutorials/recipes/recipes/amp_recipe.html
        # autocast will automatically manage which operations to run in FP16 and which ones to run in FP32.
        # eg. matmul will cast to FP16 and it's a crucial part of the whole pipeline.
        # Here the complete list of FP16 supported modules: https://docs.pytorch.org/docs/stable/amp.html#cuda-ops-that-can-autocast-to-float16
        # NOTE Switching to torch.bfloat16 for better accuracy and same efficiency of float16, more on this here: https://www.cerebras.ai/blog/to-bfloat-or-not-to-bfloat-that-is-the-question
        with autocast(
            device.type, dtype=amp_dtype_dict[config.training_hp.amp_dtype], enabled=config.training_hp.use_amp
        ):
            output_logits = model(src_sequence, tgt_input_sequence, src_mask, tgt_mask)

            if config.training_hp.loss.type == "crossentropy":
                # pred shape: [B*S, V]  (B: batch size, S: sequence length, V: vocabulary size)
                # target shape: [B*S]
                pred = output_logits.reshape(-1, output_logits.size(-1))
                target = tgt_output_label.reshape(-1)

                loss = loss_fn(pred, target)
                loss /= num_tgt_tokens  # Loss rescaling based on processed target tokens

            elif config.training_hp.loss.type == "KLdiv-labelsmoothing":
                output_log_proba = F.log_softmax(output_logits, dim=-1)

                pred = output_log_proba.reshape(-1, output_log_proba.size(-1))
                target = tgt_output_label.reshape(-1)

                loss = loss_fn(pred, target)
                loss /= num_tgt_tokens

            else:
                raise ValueError("Loss type not supported")

        # Compute gradients with mixed precision weights
        scaler.scale(loss).backward()

        # Unscaling here to prevent aggressive gradient clipping
        scaler.unscale_(optimizer)

        # Gradient clipping to stabilize training avoiding exploding gradients
        grad_norm_before_clipping = compute_grad_norm(model.parameters())
        if config.training_hp.max_gradient_norm > 0:
            nn.utils.clip_grad_norm_(model.parameters(), config.training_hp.max_gradient_norm)

        scaler.step(optimizer)
        scaler.update()
        lr_scheduler.step()

        metric = {
            "train_loss": loss.item(),
            "tokens_seen_src_cum": getattr(engine.state, "tokens_seen_src", 0),
            "tokens_seen_tgt_cum": getattr(engine.state, "tokens_seen_tgt", 0),
            "tokens_seen_tot_cum": getattr(engine.state, "tokens_seen_src", 0)
            + getattr(engine.state, "tokens_seen_tgt", 0),
            "grad_norm_before_clipping": grad_norm_before_clipping,
        }
        engine.state.metrics = metric

        return metric

    trainer = Engine(train_one_iter)

    # Initialize token seen during training counters
    @trainer.on(Events.EPOCH_STARTED)
    def init_token_seen():
        trainer.state.tokens_seen_src = 0
        trainer.state.tokens_seen_tgt = 0

    return trainer


def setup_evaluator(
    config: DictConfig | ListConfig,
    model: nn.Module,
    metrics: dict[str, Metric],
    device: torch.device,
) -> Engine | DeterministicEngine:
    """Setup an evaluator with mixed precision training support.

    Args:
        config (DictConfig | ListConfig): Project config file.
        model (nn.Module): Transformer model.
        metrics (dict[str, Metric]): Metrics to be used.
        device (torch.device): Device.

    Returns:
        tuple[Engine | DeterministicEngine, Engine | DeterministicEngine]: Evaluator objects.
    """

    # Gradient scaler is not required during evaluation.
    # https://docs.pytorch.org/tutorials/recipes/recipes/amp_recipe.html#inference-evaluation

    amp_dtype_dict = {"float32": torch.float32, "bfloat16": torch.bfloat16, "float16": torch.float16}

    @torch.no_grad()
    def eval_one_iter(engine: Engine, batch: dict[str, torch.Tensor | str]) -> tuple[torch.Tensor, torch.Tensor]:
        # NOTE See train_one_iter function for code explanation

        src_sequence = batch["src"].to(device, non_blocking=True, dtype=torch.long)
        tgt_sequence = batch["tgt"].to(device, non_blocking=True, dtype=torch.long)
        src_mask = batch["src_mask"].to(device, non_blocking=True, dtype=torch.long)
        tgt_mask = batch["tgt_mask"].to(device, non_blocking=True, dtype=torch.long)

        tgt_output_label = tgt_sequence[:, 1:]

        tgt_input_sequence = tgt_sequence[:, :-1]

        model.eval()

        with autocast(
            device.type, dtype=amp_dtype_dict[config.training_hp.amp_dtype], enabled=config.training_hp.use_amp
        ):
            output_logits = model(src_sequence, tgt_input_sequence, src_mask, tgt_mask)

        return output_logits, tgt_output_label

    test_evaluator = Engine(eval_one_iter)

    for name, metric in metrics.items():
        metric.attach(test_evaluator, name)

    return test_evaluator



from collections.abc import Callable

import torch
import torch.nn as nn
from ignite.exceptions import NotComputableError
from ignite.metrics import Metric
from ignite.metrics.metric import reinit__is_reduced, sync_all_reduce
from jaxtyping import Float


class Perplexity(Metric):
    """Computes Perplexity metric: exp( Sum(Loss) / Total_Tokens ).

    If the criterion is KLDivLabelSmoothingLoss, it returns the average loss (KL divergence) without exponentiation.

    Args:
        criterion (nn.Module): Loss criterion. The criterion must return the sum of losses.
        pad_idx (int): The padding index to ignore.
        output_transform (callable, optional): function that transforms 'x' input to update method.
        device (Union[str, torch.device], optional): device specification in case of distributed computation usage.
    """

    def __init__(
        self,
        criterion: nn.Module,
        pad_idx: int,
        output_transform: Callable = lambda x: x,
        device: str | torch.device = "cpu",
    ):
        self.criterion = (
            criterion if criterion is not None else nn.CrossEntropyLoss(ignore_index=pad_idx, reduction="sum")
        )
        self.pad_idx = pad_idx
        super().__init__(output_transform=output_transform, device=device)

    def _check_shape_and_type_consistency(self, preds: torch.Tensor, target: torch.Tensor) -> None:
        """Check shape and type consistency of input vectors."""
        if preds.dim() not in [2, 3]:
            raise ValueError(
                "Input tensor `preds` is expected to have 2 dimensions [batch_size*seq_len, vocab_size] "
                "or 3 dimensions [batch_size, seq_len, vocab_size], "
                f"but got {preds.dim()}."
            )

        if target.dim() not in [1, 2]:
            raise ValueError(
                "Input tensor `target` is expected to have 1 dimension [batch_size*seq_len] "
                "or 2 dimensions [batch_size, seq_len], "
                f"but got {target.dim()}."
            )

        if preds.dim() == 2 and target.dim() != 1:
            raise ValueError(f"If preds is 2D, target must be 1D. Got preds: {preds.shape}, target: {target.shape}")

        if preds.dim() == 3 and target.dim() != 2:
            raise ValueError(f"If preds is 3D, target must be 2D. Got preds: {preds.shape}, target: {target.shape}")

        if not preds.is_floating_point():
            raise TypeError(f"Input tensor `preds` is expected to be of floating point type but got {preds.dtype}.")

        if target.dtype != torch.int64:
            raise TypeError(f"Input tensor `target` is expected to be of a type {torch.int64} but got {target.dtype}.")

    @reinit__is_reduced
    def reset(self):
        self._sum_loss = torch.tensor(0.0, device=self._device)
        self._num_tokens = torch.tensor(0.0, device=self._device)
        super().reset()

    @reinit__is_reduced
    def update(self, output: tuple[Float[torch.Tensor, "B S D"], Float[torch.Tensor, "B S D"]]):
        y_pred, y = output

        self._check_shape_and_type_consistency(y_pred, y)

        # Reshape to (N, C) and (N,) if necessary
        if y_pred.dim() == 3:
            y_pred = y_pred.reshape(-1, y_pred.shape[-1])
            y = y.reshape(-1)

        loss = self.criterion(y_pred, y)

        # Count tokens excluding PAD tokens
        mask = y != self.pad_idx
        num_tokens = mask.sum()

        self._sum_loss += loss.item()
        self._num_tokens += num_tokens.item()

    @sync_all_reduce("_sum_loss", "_num_tokens")
    def compute(self):
        if self._num_tokens == 0:
            raise NotComputableError("Perplexity must have at least one example before it can be computed.")

        mean_loss = self._sum_loss / self._num_tokens

        if self.criterion.__class__.__name__ == "KLDivLabelSmoothingLoss":
            return mean_loss.item()

        return torch.exp(mean_loss).item()
